/**
 * Copyright (C) 2008 - Abiquo Holdings S.L. All rights reserved.
 *
 * Please see /opt/abiquo/tomcat/webapps/legal/ on Abiquo server
 * or contact contact@abiquo.com for licensing information.
 */
package com.abiquo.server.core.appslibrary;

import java.util.ArrayList;
import java.util.List;

import javax.persistence.EntityManager;

import org.springframework.beans.factory.annotation.Autowired;

import com.abiquo.model.enumerator.ConversionState;
import com.abiquo.model.enumerator.DiskFormatType;
import com.abiquo.model.enumerator.HypervisorType;
import com.abiquo.model.enumerator.OSType;
import com.abiquo.model.enumerator.StatefulInclusion;
import com.abiquo.server.core.cloud.VirtualDatacenter;
import com.abiquo.server.core.common.DefaultRepBase;
import com.abiquo.server.core.enterprise.Enterprise;
import com.abiquo.server.core.enterprise.User;
import com.abiquo.server.core.infrastructure.Repository;
import com.abiquo.server.core.pricing.CostCode;
import com.abiquo.server.core.util.FilterOptions;

@org.springframework.stereotype.Repository
public class AppsLibraryRep extends DefaultRepBase
{
    @Autowired
    private VirtualMachineTemplateDAO virtualMachineTemplateDAO;

    @Autowired
    private VirtualImageConversionDAO conversionDAO;

    @Autowired
    private CategoryDAO categoryDAO;

    public AppsLibraryRep()
    {

    }

    public AppsLibraryRep(final EntityManager em)
    {
        this.entityManager = em;
        this.virtualMachineTemplateDAO = new VirtualMachineTemplateDAO(em);
        this.categoryDAO = new CategoryDAO(em);
        this.conversionDAO = new VirtualImageConversionDAO(em);
    }

    // Category

    public void insertCategory(final Category category)
    {
        categoryDAO.persist(category);
    }

    public void updateCategory(final Category category)
    {
        categoryDAO.flush();
    }

    /**
     * If if being used by any {@link VirtualMachineTemplate} its changed to the DEFAULT category;
     */
    public void deleteCategory(final Category category)
    {
        for (VirtualMachineTemplate templ : virtualMachineTemplateDAO.findBy(category))
        {
            templ.setCategory(categoryDAO.findDefault());
        }

        categoryDAO.remove(category);
    }

    public Category getDefaultCategory()
    {
        return categoryDAO.findDefault();
    }

    public List<Category> findAllCategories(final Integer idEnterprise, final boolean onlyLocal)
    {
        List<Category> result = new ArrayList<Category>();
        if (idEnterprise != 0)
        {
            if (onlyLocal)
            {
                return categoryDAO.findLocalCategories(idEnterprise);
            }
            result.addAll(categoryDAO.findLocalCategories(idEnterprise));
        }

        result.addAll(categoryDAO.findGlobalCategories());

        return result;
    }

    public Category findCategoryById(final Integer idCategory)
    {
        return categoryDAO.findById(idCategory);
    }

    public Category findCategoryByName(final String name, final Enterprise enterprise)
    {
        return categoryDAO.findByNameAndEnterprise(name, enterprise);
    }

    public Category findByCategoryNameOrCreateNew(final String categoryName)
    {
        if (categoryName == null || categoryName.isEmpty())
        {
            return getDefaultCategory();
        }

        Category cat = findCategoryByName(categoryName, null);

        if (cat == null)
        {
            cat = new Category(categoryName);
            insertCategory(cat);
        }

        return cat;
    }

    public boolean existAnyWithName(final String name)
    {
        return categoryDAO.existsAnyWithName(name);
    }

    public boolean existAnyOtherWithName(final Category category, final String name)
    {
        return categoryDAO.existsAnyOtherWithName(category, name);
    }

    // Virtual Machine Template

    public void insertVirtualMachineTemplate(final VirtualMachineTemplate virtualMachineTemplate)
    {
        virtualMachineTemplateDAO.persist(virtualMachineTemplate);
    }

    public void updateVirtualMachineTemplate(final VirtualMachineTemplate virtualMachineTemplate)
    {
        virtualMachineTemplateDAO.flush();
    }

    public void deleteVirtualMachineTemplate(final VirtualMachineTemplate virtualMachineTemplate)
    {
        virtualMachineTemplateDAO.remove(virtualMachineTemplate);
    }

    public List<VirtualMachineTemplate> findAllVirtualMachineTemplates()
    {
        return virtualMachineTemplateDAO.findAll();
    }

    public VirtualMachineTemplate findVirtualMachineTemplateById(
        final Integer virtualMachineTemplate)
    {
        return virtualMachineTemplateDAO.findById(virtualMachineTemplate);
    }

    public VirtualMachineTemplate findVirtualMachineTemplateByName(final String name)
    {
        return virtualMachineTemplateDAO.findByName(name);
    }

    public List<VirtualMachineTemplate> findVirtualMachineTemplatesByEnterprise(
        final Enterprise enterprise)
    {
        return virtualMachineTemplateDAO.findByEnterprise(enterprise);
    }

    public List<VirtualMachineTemplate> findVirtualMachineTemplatesByEnterpriseAndRepository(
        final Enterprise enterprise,
        final com.abiquo.server.core.infrastructure.Repository repository)
    {
        return virtualMachineTemplateDAO.findByEnterpriseAndRepository(enterprise, repository);
    }

    public List<VirtualMachineTemplate> findInPrecreatedVirtualMachineTemplate(final String ovfId,
        final Enterprise enterprise,
        final com.abiquo.server.core.infrastructure.Repository repository)
    {
        return virtualMachineTemplateDAO.findInPrecreatedVirtualMachineTemplate(ovfId, enterprise,
            repository);
    }

    /**
     * Gets the list of compatible(*) virtual machine templates available in the provided enterprise
     * and repository.
     * 
     * @param category null indicate all categories (no filter)
     * @param hypervisor (*) null indicate no filter compatibles, else return templates compatibles
     *            or with compatible conversions. @see {@link VirtualImageConversionDAO}
     * @param ostype, optionally filter by operating system (null for any filter)
     * @param is64b, optionally filter by x64 architecture. It requires ''ostype'' param set to null
     *            (null for any filter)
     * @param filters The pagination and search filters
     */
    public List<VirtualMachineTemplate> findVirtualMachineTemplates(final Enterprise enterprise,
        final com.abiquo.server.core.infrastructure.Repository repository, final Category category,
        final HypervisorType hypervisor, final OSType ostype, final Boolean is64b,
        final Boolean imported, final FilterOptions filters)
    {
        return virtualMachineTemplateDAO.findBy(enterprise, repository, category, hypervisor,
            ostype, is64b, imported != null && imported.booleanValue(), filters);
    }

    public boolean existVirtualMachineTemplateWithSamePath(final Enterprise enterprise,
        final com.abiquo.server.core.infrastructure.Repository repository, final String path)
    {
        return virtualMachineTemplateDAO.existWithSamePath(enterprise, repository, path);
    }

    public VirtualMachineTemplate findVirtualMachineTemplateByPath(final Enterprise enterprise,
        final com.abiquo.server.core.infrastructure.Repository repository, final String path)
    {
        return virtualMachineTemplateDAO.findByPath(enterprise, repository, path);
    }

    public List<VirtualMachineTemplate> findStatefulVirtualMachineTemplatesByCategoryAndRepositoryAndVirutalDatacenter(
        final User user, final Enterprise enterprise, final Category category,
        final Repository repository, final VirtualDatacenter virtualdatacenter,
        final StatefulInclusion stateful, final OSType ostype, final Boolean is64b,
        final FilterOptions filters)
    {
        return virtualMachineTemplateDAO.findStatefulsByCategoryAndDatacenter(user, enterprise,
            category, repository, virtualdatacenter, stateful, ostype, is64b, filters);
    }

    public boolean isMaster(final VirtualMachineTemplate vmtemplae)
    {
        return virtualMachineTemplateDAO.isMaster(vmtemplae);
    }

    public List<VirtualMachineTemplate> findVirtualMachineTemplatesByMaster(
        final VirtualMachineTemplate master)
    {
        return virtualMachineTemplateDAO.findByMaster(master);
    }

    public List<Integer> getPersistentIds(final User user, final Enterprise enterprise,
        final Category category, final Repository repository, final StatefulInclusion stateful,
        final OSType ostype, final Boolean is64b, final FilterOptions filters)
    {
        return virtualMachineTemplateDAO.getPersistentIds(user, enterprise, category, repository,
            stateful, ostype, is64b, filters);
    }

    public List<VirtualMachineTemplate> findVirtualMachineTemplatesByCostCode(
        final CostCode costCode)
    {
        return virtualMachineTemplateDAO.findByCostCode(costCode);
    }

    /**
     * @see com.abiquo.server.core.appslibrary.VirtualImageConversionDAO#compatilbeConversions(com.abiquo.server.core.cloud.VirtualMachineTemplate,
     *      com.abiquo.model.enumerator.HypervisorType)
     */
    public List<VirtualImageConversion> compatilbeConversions(
        final VirtualMachineTemplate vmtemplate, final HypervisorType hypervisorType)
    {
        return conversionDAO.compatilbeConversions(vmtemplate, hypervisorType);
    }

    public VirtualImageConversion getConversionByTemplateAndFormat(
        final VirtualMachineTemplate virtualMachineTemplate, final DiskFormatType targetFormat)
    {
        return conversionDAO.getConversionByTemplateAndFormat(virtualMachineTemplate, targetFormat);
    }

    public List<VirtualImageConversion> getConversionsByTemplate(
        final VirtualMachineTemplate virtualMachineTemplate)
    {
        return conversionDAO.findByVirtualImage(virtualMachineTemplate);
    }

    /**
     * @param hypervisorType, optionally restrict to compatible conversions
     * @param state, optionally restrict to some conversion state
     */
    public List<VirtualImageConversion> getConversionsByTemplate(
        final VirtualMachineTemplate virtualMachineTemplate, final HypervisorType hypervisorType,
        final ConversionState state)
    {
        return conversionDAO.getConversionsOptionalHypervisorTypeAndState(virtualMachineTemplate,
            hypervisorType, state);
    }

    public boolean isVirtualMachineTemplateConverted(final VirtualMachineTemplate vmtemplate,
        final DiskFormatType targetType)
    {
        return conversionDAO.isConverted(vmtemplate, targetType);
    }

    public void addConversion(final VirtualImageConversion conversion)
    {
        conversionDAO.persist(conversion);
    }

    public boolean existDuplicatedConversion(final VirtualImageConversion conversion)
    {
        return conversionDAO.existDuplicatedConversion(conversion);
    }

    public void updateConversion(final VirtualImageConversion conversion)
    {
        conversionDAO.flush();
    }
}
