/**
 * Copyright (C) 2008 - Abiquo Holdings S.L. All rights reserved.
 *
 * Please see /opt/abiquo/tomcat/webapps/legal/ on Abiquo server
 * or contact contact@abiquo.com for licensing information.
 */
package com.abiquo.server.core.cloud;

import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.persistence.EntityManager;
import javax.persistence.LockModeType;
import javax.persistence.TypedQuery;

import org.apache.commons.lang.StringUtils;
import org.hibernate.Criteria;
import org.hibernate.Query;
import org.hibernate.criterion.Criterion;
import org.hibernate.criterion.Order;
import org.hibernate.criterion.Restrictions;
import org.hibernate.sql.JoinFragment;
import org.springframework.stereotype.Repository;

import com.abiquo.server.core.appslibrary.VirtualImageConversion;
import com.abiquo.server.core.appslibrary.VirtualMachineTemplate;
import com.abiquo.server.core.cloud.VirtualMachine.OrderByEnum;
import com.abiquo.server.core.common.persistence.DefaultDAOBase;
import com.abiquo.server.core.common.persistence.JPAConfiguration;
import com.abiquo.server.core.enterprise.Enterprise;
import com.abiquo.server.core.enterprise.User;
import com.abiquo.server.core.infrastructure.Datacenter;
import com.abiquo.server.core.task.enums.TaskType;
import com.abiquo.server.core.util.FilterOptions;
import com.abiquo.server.core.util.PagedList;
import com.google.common.base.Function;
import com.google.common.base.Strings;
import com.google.common.collect.Multimaps;
import com.softwarementors.bzngine.entities.PersistentEntity;

@Repository("jpaVirtualMachineDAO")
@SuppressWarnings("unchecked")
public class VirtualMachineDAO extends DefaultDAOBase<Integer, VirtualMachine>
{
    public static final String BY_VAPP_AND_ID = "SELECT nvi.virtualMachine "
        + "FROM NodeVirtualImage nvi "
        + "WHERE nvi.virtualAppliance.id = :vapp_id AND nvi.virtualMachine.id = :vm_id";

    private static Criterion equalName(final String name)
    {
        assert !StringUtils.isEmpty(name);

        return Restrictions.eq(Datacenter.NAME_PROPERTY, name);
    }

    private static Criterion sameEnterprise(final Enterprise enterprise)
    {
        assert enterprise != null;

        return Restrictions.eq(VirtualMachine.ENTERPRISE_PROPERTY, enterprise);
    }

    private static Criterion sameHypervisor(final Hypervisor hypervisor)
    {
        assert hypervisor != null;

        return Restrictions.eq(VirtualMachine.HYPERVISOR_PROPERTY, hypervisor);
    }

    private static Criterion sameUser(final User user)
    {
        return Restrictions.eq(VirtualMachine.USER_PROPERTY, user);
    }

    public VirtualMachineDAO()
    {
        super(VirtualMachine.class);
    }

    public VirtualMachineDAO(final EntityManager entityManager)
    {
        super(VirtualMachine.class, entityManager);
    }

    public List<VirtualMachine> getNotManagedVirtualMachines(final Hypervisor hypervisor)
    {
        Criteria criteria = createCriteria(sameHypervisor(hypervisor), notManaged());
        return getResultList(criteria);
    }

    public boolean existsAnyWithName(final String name)
    {
        assert !StringUtils.isEmpty(name);

        return this.existsAnyByCriterions(equalName(name));
    }

    public VirtualMachine findByIdByVirtualApp(final VirtualAppliance vapp, final Integer vmId)
    {
        Query finalQuery = getSession().createQuery(BY_VAPP_AND_ID);
        finalQuery.setParameter("vapp_id", vapp.getId());
        finalQuery.setParameter("vm_id", vmId);

        return (VirtualMachine) finalQuery.uniqueResult();
    }

    public VirtualMachine findByName(final String name)
    {
        return findUniqueByProperty(VirtualMachine.NAME_PROPERTY, name);
    }

    public VirtualMachine findByUUID(final String uuid)
    {
        return findUniqueByProperty(VirtualMachine.UUID_PROPERTY, uuid);
    }

    public List<VirtualMachine> findManagedVirtualMachines(final Hypervisor hypervisor)
    {
        assert hypervisor != null;
        assert isManaged2(hypervisor);

        Criteria criteria = createCriteria(sameHypervisor(hypervisor), managed());
        criteria.addOrder(Order.asc(VirtualMachine.NAME_PROPERTY));
        List<VirtualMachine> result = getResultList(criteria);
        return result;
    }

    public VirtualMachine findVirtualMachineByHypervisor(final Hypervisor hypervisor,
        final Integer virtualMachineId)
    {
        Criteria criteria =
            createCriteria(sameHypervisor(hypervisor), Restrictions.eq(
                PersistentEntity.ID_PROPERTY, virtualMachineId));
        return (VirtualMachine) criteria.uniqueResult();
    }

    public List<VirtualMachine> findVirtualMachines(final Hypervisor hypervisor)
    {
        assert hypervisor != null;
        assert isManaged2(hypervisor);

        Criteria criteria = createCriteria(sameHypervisor(hypervisor));
        criteria.addOrder(Order.asc(VirtualMachine.NAME_PROPERTY));
        List<VirtualMachine> result = getResultList(criteria);
        return result;
    }

    // without hypervisor
    public List<VirtualMachine> findVirtualMachinesNotAllocatedCompatibleHypervisor(
        final Hypervisor hypervisor)
    {
        Criteria criteria = createCriteria();
        criteria.createAlias(VirtualMachine.VIRTUAL_MACHINE_TEMPLATE_PROPERTY, "template");
        criteria.createAlias(VirtualMachine.VIRTUAL_IMAGE_CONVERSION_PROPERTY, "conversion",
            JoinFragment.LEFT_OUTER_JOIN);
        criteria.createAlias("conversion."
            + VirtualImageConversion.VIRTUAL_MACHINE_TEMPLATE_PROPERTY, "masterconversion",
            JoinFragment.LEFT_OUTER_JOIN);
        criteria.add(Restrictions.and(Restrictions.isNull(VirtualMachine.HYPERVISOR_PROPERTY),
            Restrictions.or(Restrictions.or(Restrictions.and(Restrictions
                .isNull(VirtualMachine.VIRTUAL_IMAGE_CONVERSION_PROPERTY), Restrictions.in(
                "template." + VirtualMachineTemplate.DISKFORMAT_TYPE_PROPERTY, Arrays
                    .asList(hypervisor.getType().compatibilityTable))), Restrictions.and(
                Restrictions.eqProperty("masterconversion." + PersistentEntity.ID_PROPERTY,
                    "template." + PersistentEntity.ID_PROPERTY), Restrictions.in("conversion."
                    + VirtualImageConversion.TARGET_TYPE_PROPERTY, Arrays.asList(hypervisor
                    .getType().compatibilityTable)))), Restrictions.eq("template."
                + VirtualMachineTemplate.STATEFUL_PROPERTY, Boolean.TRUE))));
        criteria.addOrder(Order.asc(VirtualMachine.NAME_PROPERTY));
        List<VirtualMachine> result = getResultList(criteria);
        return result;
    }

    public List<VirtualMachine> findVirtualMachinesByDatacenter(final Integer datacenterId)
    {
        List<VirtualMachine> vmList = null;
        TypedQuery<VirtualMachine> query =
            getEntityManager().createNamedQuery("VIRTUAL_MACHINE.BY_DC", VirtualMachine.class);
        query.setParameter("datacenterId", datacenterId);
        vmList = query.getResultList();

        return vmList;
    }

    public List<VirtualMachine> findVirtualMachinesByEnterprise(final Enterprise enterprise)
    {
        TypedQuery<VirtualMachine> query =
            getEntityManager().createNamedQuery("VIRTUAL_MACHINE.BY_ENTERPRISE",
                VirtualMachine.class);
        query.setParameter("enterprise", enterprise);
        query.setParameter("filterLike", "%");
        return query.getResultList();
    }

    public List<VirtualMachine> findVirtualMachinesByEnterprise(final Enterprise enterprise,
        final FilterOptions filters)
    {
        Query query = getSession().getNamedQuery("VIRTUAL_MACHINE.BY_ENTERPRISE");

        // Add order filter to the query
        VirtualMachine.OrderByEnum orderByEnum =
            VirtualMachine.OrderByEnum.fromValue(filters.getOrderBy());
        String orderBy = defineOrderBy(orderByEnum.getColumnHQL(), filters.getAsc());
        Query queryWithOrder = getSession().createQuery(query.getQueryString() + orderBy);

        // Add the query parameters
        Map<String, Object> params = new HashMap<String, Object>();
        params.put("enterprise", enterprise);
        params.put("filterLike", filters.getFilter().isEmpty() ? "%" : "%" + filters.getFilter()
            + "%");
        queryWithOrder.setProperties(params);

        // Get the total number of results
        Integer size = count(queryWithOrder, params).intValue();

        // Limit 0 means no size filter
        int limit = filters.getLimit();
        int startwith = filters.getStartwith();
        if (limit == 0)
        {
            limit = size;
            startwith = 0;
        }

        queryWithOrder.setFirstResult(startwith);
        queryWithOrder.setMaxResults(limit);

        PagedList<VirtualMachine> vms = new PagedList<VirtualMachine>(queryWithOrder.list());
        vms.setTotalResults(size);
        vms.setPageSize(limit > size ? size : limit);
        vms.setCurrentElement(startwith);

        return vms;
    }

    public List<VirtualMachine> findVirtualMachinesByVirtualDatacenters(final Integer... ids)
    {
        TypedQuery<VirtualMachine> query =
            getEntityManager().createNamedQuery("VIRTUAL_MACHINE.BY_VDC", VirtualMachine.class);
        query.setParameter("vdcIds", Arrays.asList(ids));
        query.setParameter("filterLike", "%");
        return query.getResultList();
    }

    public List<VirtualMachine> findVirtualMachinesByVirtualDatacenters(
        final FilterOptions filters, final Integer... ids)
    {
        Query query = getSession().getNamedQuery("VIRTUAL_MACHINE.BY_VDC");

        // Add order filter to the query
        VirtualMachine.OrderByEnum orderByEnum =
            VirtualMachine.OrderByEnum.fromValue(filters.getOrderBy());
        String orderBy = defineOrderBy(orderByEnum.getColumnHQL(), filters.getAsc());
        Query queryWithOrder = getSession().createQuery(query.getQueryString() + orderBy);

        // Add the query parameters
        Map<String, Object> params = new HashMap<String, Object>();
        params.put("vdcIds", Arrays.asList(ids));
        params.put("filterLike", filters.getFilter().isEmpty() ? "%" : "%" + filters.getFilter()
            + "%");
        queryWithOrder.setProperties(params);

        // Get the total number of results
        Integer size = count(queryWithOrder, params).intValue();

        // Limit 0 means no size filter
        int limit = filters.getLimit();
        int startwith = filters.getStartwith();
        if (limit == 0)
        {
            limit = size;
            startwith = 0;
        }

        queryWithOrder.setFirstResult(startwith);
        queryWithOrder.setMaxResults(limit);

        PagedList<VirtualMachine> vms = new PagedList<VirtualMachine>(queryWithOrder.list());
        vms.setTotalResults(size);
        vms.setPageSize(limit > size ? size : limit);
        vms.setCurrentElement(startwith);

        return vms;
    }

    public List<VirtualMachine> findVirtualMachinesByUser(final Enterprise enterprise,
        final User user)
    {
        Criteria criteria = createCriteria(sameUser(user)).add(sameEnterprise(enterprise));
        criteria.addOrder(Order.asc(VirtualMachine.NAME_PROPERTY));

        return criteria.list();
    }

    public List<VirtualMachine> findVirtualMachinesByVirtualAppliance(final Integer vappId,
        Integer startwith, Integer limit, final String filter, final OrderByEnum orderby,
        final boolean asc)
    {
        // List<VirtualMachine> vmList = null;

        // TypedQuery<VirtualMachine> query =
        // getEntityManager().createNamedQuery("VIRTUAL_MACHINE.BY_VAPP", VirtualMachine.class);
        // query.setParameter("vapp_id", vappId);
        // vmList = query.getResultList();
        String orderBy = defineOrderBy(orderby.getColumnHQL(), asc);
        Query query = getSession().getNamedQuery("VIRTUAL_MACHINE.BY_VAPP");

        String req = query.getQueryString() + orderBy;
        // Add order filter to the query
        Query queryWithOrder = getSession().createQuery(req);

        Map<String, Object> params = new HashMap<String, Object>();
        params.put("vapp_id", vappId);
        params.put("filterLike", filter.isEmpty() ? "%" : "%" + filter + "%");
        queryWithOrder.setProperties(params);

        Integer size = count(queryWithOrder, params).intValue();

        // Limit 0 means no size filter
        if (limit == 0)
        {
            limit = size;
            startwith = 0;
        }

        queryWithOrder.setFirstResult(startwith);
        queryWithOrder.setMaxResults(limit);

        PagedList<VirtualMachine> vmList = new PagedList<VirtualMachine>(queryWithOrder.list());
        vmList.setTotalResults(size);
        vmList.setPageSize(limit > size ? size : limit);
        vmList.setCurrentElement(startwith);

        return vmList;
    }

    public List<VirtualMachine> findByVirtualMachineTemplate(final Integer virtualMachineTemplateId)
    {
        List<VirtualMachine> vmList = null;
        TypedQuery<VirtualMachine> query =
            getEntityManager().createNamedQuery("VIRTUAL_MACHINE.BY_VMT", VirtualMachine.class);
        query.setParameter("virtualMachineTplId", virtualMachineTemplateId);
        vmList = query.getResultList();

        return vmList;
    }

    public List<VirtualMachine> findVirtualInfrastructureCheckCandidates()
    {
        Query query = getSession().createQuery(VirtualMachine.VI_CHECK_CANDIDATES);
        return query.list();
    }

    public boolean hasVirtualMachineTemplate(final Integer virtualMachineTemplateId)
    {
        TypedQuery<Long> query =
            getEntityManager().createNamedQuery("VIRTUAL_MACHINE.HAS_VMT", Long.class);
        query.setParameter("virtualMachineTplId", virtualMachineTemplateId);
        return query.getResultList().get(0) > 0;
    }

    public void updateVirtualMachineState(final Integer vmachineId, final VirtualMachineState state)
    {
        VirtualMachine vmachine = findById(vmachineId);

        vmachine.setState(state);

        flush();
    }

    /**
     * Sets the {@link VirtualMachine#setState(VirtualMachineState)} to
     * {@link VirtualMachineState#UNKNOWN}.
     * 
     * @param vmachineId id void
     */
    public void unknownState(final Integer vmachineId)
    {
        VirtualMachine vmachine = findById(vmachineId);

        vmachine.setState(VirtualMachineState.UNKNOWN);
    }

    /**
     * Sets the {@link VirtualMachine#setState(VirtualMachineState)} to the given
     * {@link VirtualMachineState}
     * 
     * @param virtualMachineId the id of the virtual machine to update
     * @param state the state to set
     */
    public void setState(final Integer virtualMachineId, final VirtualMachineState state)
    {
        VirtualMachine virtualMachine = findById(virtualMachineId);
        virtualMachine.setState(state);
    }

    private Criterion managed()
    {
        return Restrictions.eq(VirtualMachine.ID_TYPE_PROPERTY, VirtualMachine.MANAGED);
    }

    private Criterion notManaged()
    {
        return Restrictions.eq(VirtualMachine.ID_TYPE_PROPERTY, VirtualMachine.NOT_MANAGED);
    }

    private Criteria temporalVirtualMachine(final Integer vmachineId)
    {
        return createCriteria(Restrictions.eq(VirtualMachine.TEMPORAL_PROPERTY, vmachineId));
    }

    public VirtualMachine findBackup(final VirtualMachine vmachine)
    {

        try
        {
            JPAConfiguration.enableOnlyTemporalFilters(getEntityManager());

            return getSingleResult(temporalVirtualMachine(vmachine.getId()));
        }
        finally
        {
            JPAConfiguration.enableDefaultFilters(getEntityManager());
        }
    }

    public void refreshLock(final VirtualMachine vm)
    {
        // We force the refresh from database (PESSIMISTIC) and increment the version of the object
        // (INCREMENT) to ensure that no other code will be using the "unlocked" object.
        getEntityManager().refresh(vm, LockModeType.PESSIMISTIC_FORCE_INCREMENT);
    }

    public void refresh(final VirtualMachine vm)
    {
        getEntityManager().refresh(vm);
    }

    public void detachHypervisor(final VirtualMachine vm)
    {
        getEntityManager().detach(vm.getHypervisor());
    }

    public void detachVirtualMachine(final VirtualMachine vm)
    {
        getEntityManager().detach(vm);
    }

    private String defineOrderBy(final String orderBy, final Boolean asc)
    {
        StringBuilder queryString = new StringBuilder();

        queryString.append(" order by ");
        queryString.append(orderBy);
        queryString.append(" ");

        if (asc)
        {
            queryString.append("asc");
        }
        else
        {
            queryString.append("desc");
        }

        return queryString.toString();
    }

    public VirtualMachine get(final Integer id)
    {
        return getEntityManager().find(VirtualMachine.class, id);
    }

    /**
     * Group virtual machines to layers
     * 
     * @param vappId, virtual appliance for all the returned virtual machines
     * @return map indexed by the Layer name to all the virtual machines that belongs to this layer
     */
    public Map<String, Collection<VirtualMachine>> groupByLayer(final Integer vappId)
    {

        TypedQuery<VirtualMachine> query =
            getEntityManager().createNamedQuery("VIRTUAL_MACHINE.LAYER", VirtualMachine.class);
        query.setParameter("vapp_id", vappId);
        List<VirtualMachine> vms = query.getResultList();

        return Multimaps.index(vms, new Function<VirtualMachine, String>()
        {
            @Override
            public String apply(final VirtualMachine input)
            {
                return input.getLayer();
            }
        }).asMap();
    }

    /**
     * Get all the virtual machines in the layer
     * 
     * @param vappId, virtual appliance for all the returned virtual machines
     * @param layerName, name of the layer for all the returned virtual machines
     * @return all the virtual machines with layerName in the vapp
     */
    public List<VirtualMachine> findByLayer(final Integer vappId, final String layerName)
    {
        Query query = getSession().createQuery(VirtualMachine.LAYER_NAME);
        query.setParameter("vapp_id", vappId);
        query.setParameter("layerName", layerName);
        return query.list();
    }

    /**
     * Get all the virtual machines UUIDs in the layer
     * 
     * @param vappId, virtual appliance for all the returned virtual machines
     * @param layerName, name of the layer for all the returned virtual machines
     * @return all the virtual machine UUIDs with layerName in the vapp
     */
    public List<String> findNamesByLayer(final Integer vappId, final String layerName)
    {
        Query query = getSession().createQuery(VirtualMachine.LAYER_NAMES);
        query.setParameter("vapp_id", vappId);
        query.setParameter("layerName", layerName);
        return query.list();
    }

    /**
     * Number of virtual machines without Hypervisor in provided ''layerName'' and the ''vapp''
     * 
     * @param vappId, virtual appliance where find virtual machines
     * @param layerName, name of the layer for all the counted virtual machines
     * @return the number of virtual machines in the vapp with layer = layerName , and it needs to
     *         be deployed (0 if any)
     */
    public Long numWithoutHypervisorInLayer(final Integer vappId, final String layerName)
    {
        TypedQuery<Long> query =
            getEntityManager().createNamedQuery("VIRTUAL_MACHINE.LAYER_COUNT", Long.class);
        query.setParameter("vapp_id", vappId);
        query.setParameter("layerName", layerName);

        return query.getResultList().get(0);
    }

    /**
     * Get the already used machines in the layer
     * 
     * @param vappId, virtual appliance where find virtual machines
     * @param layerName, name of the layer for all the counted virtual machines
     * @return the machine ID used in the layers
     */
    public List<Integer> machineIdsUsedInLayer(final Integer vappId, final String layerName)
    {
        TypedQuery<Integer> query =
            getEntityManager().createNamedQuery("VIRTUAL_MACHINE.LAYER_MACHINES", Integer.class);
        query.setParameter("vapp_id", vappId);
        query.setParameter("layerName", layerName);

        return query.getResultList();
    }

    /**
     * Get all the virtual machines in the virtual appliance without any layer
     * 
     * @param vappId, virtual appliance for all the returned virtual machines
     * @return all the virtual machines without layerName in the vapp
     */
    public List<VirtualMachine> findByNoLayer(final Integer vappId)
    {
        Query query = getSession().createQuery(VirtualMachine.NO_LAYER);
        query.setParameter("vapp_id", vappId);
        return query.list();
    }

    public void setVirtualMachineState(final Integer vmId, final VirtualMachineState state)
    {
        VirtualMachine vmachine = findById(vmId);

        vmachine.setState(state);
    }

    /**
     * Checks if there is more than one {@link VirtualMachine} in the same anti affinity group (
     * {@link VirtualMachine#getLayer()}).<br />
     * <p>
     * <b>If the layer layer is not set, no rule is broken.</b>
     * </p>
     * <p>
     * This method check disregarding the state of the {@link VirtualMachine}. <br>
     * <b>If the state is {@link VirtualMachineState#UNKNOWN} and has {@link Hypervisor} assigned it
     * does not necessarily means that it exists in the actual hypervisor</b>
     * </p>
     * <p>
     * <b>If the state is {@link VirtualMachineState#LOCKED} the action in progress might be an
     * {@link TaskType#UNDEPLOY}. Nonetheless in the moment of the execution it belongs to the
     * hypervisor</b>
     * </p>
     * 
     * @param hypervisor hypervisor to check that the rule is not broken.
     * @param virtualMachine anti affinity rule holder. <b>If the layer does not exists no rule is
     *            broken</b>
     * @return true if there is two or more {@link VirtualMachine} with same layer. False otherwise
     *         or if the layer is null or empty.
     */
    public boolean isAntiaffinityBroken(final Hypervisor hypervisor,
        final VirtualMachine virtualMachine)
    {
        if (Strings.isNullOrEmpty(virtualMachine.getLayer()))
        {
            return Boolean.FALSE;
        }

        TypedQuery<Long> query =
            getEntityManager().createNamedQuery("VIRTUAL_MACHINE.HYPERVISOR_LAYER_COUNT",
                Long.class);
        query.setParameter("hyp_id", hypervisor.getId());
        query.setParameter("vm_id", virtualMachine.getId());
        query.setParameter("layerName", virtualMachine.getLayer());
        // A count always return a item
        return query.getResultList().get(0) > 1;
    }
}
