/**
 * Copyright (C) 2008 - Abiquo Holdings S.L. All rights reserved.
 *
 * Please see /opt/abiquo/tomcat/webapps/legal/ on Abiquo server
 * or contact contact@abiquo.com for licensing information.
 */
package com.abiquo.server.core.cloud;

import java.util.Collection;
import java.util.List;
import java.util.Map;

import javax.persistence.EntityManager;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Repository;

import com.abiquo.server.core.cloud.VirtualMachine.OrderByEnum;
import com.abiquo.server.core.cloud.chef.RunlistElement;
import com.abiquo.server.core.cloud.chef.RunlistElementDAO;
import com.abiquo.server.core.common.DefaultRepBase;
import com.abiquo.server.core.common.persistence.JPAConfiguration;
import com.abiquo.server.core.enterprise.Enterprise;
import com.abiquo.server.core.enterprise.User;
import com.abiquo.server.core.infrastructure.management.RasdManagement;
import com.abiquo.server.core.infrastructure.management.RasdManagementDAO;
import com.abiquo.server.core.util.FilterOptions;

@Repository
public class VirtualMachineRep extends DefaultRepBase
{
    private static final Logger LOGGER = LoggerFactory.getLogger(VirtualMachineRep.class);

    @Autowired
    private VirtualMachineDAO dao;

    @Autowired
    private RasdManagementDAO rasdDao;

    @Autowired
    private RunlistElementDAO chefDao;

    @Autowired
    private NodeVirtualImageDAO nodeVirtualImageDAO;

    public VirtualMachineRep()
    {

    }

    public VirtualMachineRep(final EntityManager entityManager)
    {
        setEntityManager(entityManager);
        this.dao = new VirtualMachineDAO(entityManager);
        this.rasdDao = new RasdManagementDAO(entityManager);
        this.chefDao = new RunlistElementDAO(entityManager);
        this.nodeVirtualImageDAO = new NodeVirtualImageDAO(entityManager);
    }

    public List<VirtualMachine> findByHypervisor(final Hypervisor hypervisor)
    {
        return dao.findVirtualMachines(hypervisor);
    }

    public List<VirtualMachine> findManagedByHypervisor(final Hypervisor hypervisor)
    {
        return dao.findManagedVirtualMachines(hypervisor);
    }

    public List<VirtualMachine> findVirtualMachinesNotAllocatedCompatibleHypervisor(
        final Hypervisor hypervisor)
    {
        return dao.findVirtualMachinesNotAllocatedCompatibleHypervisor(hypervisor);
    }

    public List<VirtualMachine> findByEnterprise(final Enterprise enterprise)
    {
        return dao.findVirtualMachinesByEnterprise(enterprise);
    }

    public List<VirtualMachine> findByEnterprise(final Enterprise enterprise,
        final FilterOptions filters)
    {
        return dao.findVirtualMachinesByEnterprise(enterprise, filters);
    }

    public List<VirtualMachine> findVirtualMachinesByUser(final Enterprise enterprise,
        final User user)
    {
        return dao.findVirtualMachinesByUser(enterprise, user);
    }

    public List<VirtualMachine> findVirtualMachinesByVirtualAppliance(final Integer vappId,
        final Integer startwith, final Integer limit, final String filter,
        final OrderByEnum orderByEnum, final Boolean descOrAsc)
    {
        return dao.findVirtualMachinesByVirtualAppliance(vappId, startwith, limit, filter,
            orderByEnum, descOrAsc);
    }

    public List<VirtualMachine> findByVirtualDatacenters(final Integer... vdcIds)
    {
        return dao.findVirtualMachinesByVirtualDatacenters(vdcIds);
    }

    public List<VirtualMachine> findByVirtualDatacenters(final FilterOptions filters,
        final Integer... vdcIds)
    {
        return dao.findVirtualMachinesByVirtualDatacenters(filters, vdcIds);
    }

    public List<VirtualMachine> findByVirtualMachineTemplate(final Integer virtualMachineTemplateId)
    {
        return dao.findByVirtualMachineTemplate(virtualMachineTemplateId);
    }

    public boolean hasVirtualMachineTemplate(final Integer virtualMachineTemplateId)
    {
        return dao.hasVirtualMachineTemplate(virtualMachineTemplateId);
    }

    public VirtualMachine findByUUID(final String uuid)
    {
        return dao.findByUUID(uuid);
    }

    public VirtualMachine findByName(final String name)
    {
        return dao.findByName(name);
    }

    public VirtualMachine findVirtualMachineById(final Integer vmId)
    {
        return dao.get(vmId);
    }

    /**
     * Sets the {@link VirtualMachine#setState(VirtualMachineState)} to
     * {@link VirtualMachineState#UNKNOWN}.
     * 
     * @param vmId id void
     */
    public void setVirtualMachineToUnknown(final Integer vmId)
    {
        dao.unknownState(vmId);
    }

    /**
     * Sets the {@link VirtualMachine#setState(VirtualMachineState)} to the given
     * {@link VirtualMachineState}
     * 
     * @param virtualMachineId the id of the virtual machine to update
     * @param state the state to set
     */
    public void setVirtualMachineState(final Integer vmId, final VirtualMachineState state)
    {
        dao.setVirtualMachineState(vmId, state);
    }

    public VirtualMachine findVirtualMachineByHypervisor(final Hypervisor hypervisor,
        final Integer virtualMachineId)
    {
        return dao.findVirtualMachineByHypervisor(hypervisor, virtualMachineId);
    }

    public void update(final VirtualMachine vm)
    {
        dao.flush();
    }

    public void refreshLock(final VirtualMachine vm)
    {
        dao.refreshLock(vm);
    }

    public List<RasdManagement> findRasdManagementByVirtualMachine(
        final VirtualMachine virtualMachine)
    {
        return rasdDao.findByVirtualMachine(virtualMachine);
    }

    /**
     * Delete a {@link VirtualMachine}.
     * 
     * @param virtualMachine to delete. void
     */
    public void deleteVirtualMachine(final VirtualMachine virtualMachine)
    {
        this.dao.remove(virtualMachine);
    }

    /**
     * Persists a {@link VirtualMachine}.
     * 
     * @param virtualMachine to create. void
     */
    public VirtualMachine createVirtualMachine(final VirtualMachine virtualMachine)
    {
        this.dao.persist(virtualMachine);
        return virtualMachine;
    }

    public void insert(final VirtualMachine virtualMachine)
    {
        this.dao.persist(virtualMachine);
        this.dao.flush();
    }

    public RunlistElement findRunlistElementById(final Integer id)
    {
        return chefDao.findById(id);
    }

    public void insertRunlistElement(final RunlistElement runlistElement)
    {
        chefDao.persist(runlistElement);
    }

    public void updateRunlistElements()
    {
        chefDao.flush();
    }

    public void deleteRunlistElement(final RunlistElement runlistElement)
    {
        chefDao.remove(runlistElement);
    }

    public List<RunlistElement> findRunlistByVirtualMachine(final VirtualMachine virtualMachine)
    {
        return chefDao.findByVirtualMachine(virtualMachine);
    }

    public void clearVirtualMachineRunlist(final VirtualMachine virtualMachine)
    {
        chefDao.clearVirtualMachineRunlist(virtualMachine);
    }

    public NodeVirtualImage insertNodeVirtualImage(final NodeVirtualImage nodeVirtualImage)
    {
        nodeVirtualImageDAO.persist(nodeVirtualImage);
        return nodeVirtualImage;
    }

    public NodeVirtualImage findNodeVirtualImageByVm(final VirtualMachine virtualMachine)
    {
        return nodeVirtualImageDAO.findByVirtualMachine(virtualMachine);
    }

    public List<NodeVirtualImage> findNodeVirtualImagesByVirtualAppliance(
        final VirtualAppliance virtualAppliance)
    {
        return nodeVirtualImageDAO.findByVirtualAppliance(virtualAppliance);
    }

    public void updateNodeVirtualImage(final NodeVirtualImage nodeVirtualImage)
    {
        nodeVirtualImageDAO.flush();
    }

    public void deleteNodeVirtualImage(final NodeVirtualImage nodeVirtualImage)
    {
        nodeVirtualImageDAO.remove(nodeVirtualImage);
    }

    public VirtualMachine findBackup(final VirtualMachine vmachine)
    {
        return dao.findBackup(vmachine);
    }

    public void refresh(final VirtualMachine vmachine)
    {
        dao.refresh(vmachine);
    }

    public void detachHypervisor(final VirtualMachine vm)
    {
        dao.detachHypervisor(vm);
    }

    public void detachVirtualMachine(final VirtualMachine vm)
    {
        dao.detachVirtualMachine(vm);
    }

    public void setDefaultFilters()
    {
        JPAConfiguration.enableDefaultFilters(this.entityManager);
    }

    public boolean existsVirtualMachineFromTemplate(final Integer virtualMachineTemplateId)
    {
        return dao.hasVirtualMachineTemplate(virtualMachineTemplateId);
    }

    public List<VirtualMachine> findVirtualInfrastructureCheckCandidates()
    {
        return dao.findVirtualInfrastructureCheckCandidates();
    }

    /**
     * Group virtual machines to layers
     * 
     * @param vappId, virtual appliance for all the returned virtual machines
     * @return map indexed by the Layer name to all the virtual machines that belongs to this layer
     */
    public Map<String, Collection<VirtualMachine>> groupVirtualMachinesByLayer(final Integer vappId)
    {
        return dao.groupByLayer(vappId);
    }

    /**
     * Get all the virtual machines of the same virtual appliance with the provided layer
     * 
     * @param vappId, virtual appliance for all the returned virtual machines
     * @param layerName, name of the layer for all the returned virtual machines
     * @return all the virtual machines with layerName in the vapp
     */
    public List<VirtualMachine> findVirtualMachinesByLayer(final Integer vappId,
        final String layerName)
    {
        return dao.findByLayer(vappId, layerName);
    }

    /**
     * Get all the virtual machines in the virtual appliance without any layer
     * 
     * @param vappId, virtual appliance for all the returned virtual machines
     * @return all the virtual machines without layerName in the vapp
     */
    public List<VirtualMachine> findVirtualMachinesByNoLayer(final Integer vappId)
    {
        return dao.findByNoLayer(vappId);
    }

    /**
     * Number of virtual machines without Hypervisor in provided ''layerName'' and the ''vapp''
     * 
     * @param vappId, virtual appliance where find virtual machines
     * @param layerName, name of the layer for all the counted virtual machines
     * @return the number of virtual machines in the vapp with layer = layerName , and it needs to
     *         be deployed (0 if any)
     */
    public Long numVirtualMachinesWithoutHypervisorInLayer(final Integer vappId,
        final String layerName)
    {
        return dao.numWithoutHypervisorInLayer(vappId, layerName);
    }

    /**
     * Checks if there is more than one {@link VirtualMachine} in the same anti affinity group (
     * {@link VirtualMachine#getLayer()}).
     * 
     * @see VirtualMachineDAO#isAntiaffinityBroken(Hypervisor, VirtualMachine)
     * @param hypervisor hypervisor to check that the rule is not broken.
     * @param virtualMachine virtual machine holding the layer.
     * @return true if there is two or more {@link VirtualMachine} with same layer. False otherwise.
     */
    public boolean isAntiaffinityBroken(final Hypervisor hypervisor,
        final VirtualMachine virtualMachine)
    {
        LOGGER
            .debug(
                "isAntiaffinityBroken: Checking anti affinity rule for virtual machine {}, layer {} in hypervisor {}",
                new Object[] {virtualMachine.getName(), virtualMachine.getLayer(),
                hypervisor.getIp()});
        return dao.isAntiaffinityBroken(hypervisor, virtualMachine);
    }
}
