/**
 * Copyright (C) 2008 - Abiquo Holdings S.L. All rights reserved.
 *
 * Please see /opt/abiquo/tomcat/webapps/legal/ on Abiquo server
 * or contact contact@abiquo.com for licensing information.
 */
package com.abiquo.server.core.enterprise;

import java.util.List;

import javax.persistence.EntityManager;
import javax.persistence.PersistenceException;

import org.hibernate.Criteria;
import org.hibernate.criterion.Projections;
import org.hibernate.criterion.Restrictions;
import org.hibernate.transform.AliasToBeanConstructorResultTransformer;
import org.springframework.stereotype.Repository;

import com.abiquo.server.core.common.persistence.DefaultDAOBase;
import com.abiquo.server.core.infrastructure.Datacenter;
import com.abiquo.server.core.infrastructure.storage.Tier;
import com.softwarementors.bzngine.entities.PersistentEntity;

@Repository("jpaAllowedTierDAO")
public class AllowedTierDAO extends DefaultDAOBase<Integer, AllowedTier>
{
    private final static String LIMIT_ALIAS = "limit";

    public AllowedTierDAO()
    {
        super(AllowedTier.class);
    }

    public AllowedTierDAO(final EntityManager entityManager)
    {
        super(AllowedTier.class, entityManager);
    }

    public AllowedTier findAllowedTierByEnterpriseAndTier(final Enterprise enterprise,
        final Tier tier)
    {
        Criteria criteria =
            createCriteria()
                .createAlias(AllowedTier.DATACENTER_LIMIT_PROPERTY, LIMIT_ALIAS)
                .add(
                    Restrictions.eq(LIMIT_ALIAS + "." + DatacenterLimits.ENTERPRISE_PROPERTY + "."
                        + PersistentEntity.ID_PROPERTY, enterprise.getId()))
                .add(
                    Restrictions.eq(AllowedTier.TIER_PROPERTY + "." + PersistentEntity.ID_PROPERTY,
                        tier.getId()));
        return (AllowedTier) criteria.uniqueResult();
    }

    @SuppressWarnings("unchecked")
    public List<AllowedTier> getAllowedTiersByTier(final Tier tier)
    {
        Criteria criteria =
            createCriteria(Restrictions.eq(AllowedTier.TIER_PROPERTY + "."
                + PersistentEntity.ID_PROPERTY, tier.getId()));
        return criteria.list();
    }

    @SuppressWarnings("unchecked")
    public List<Tier> getTiersByLimit(final DatacenterLimits limit)
    {
        Criteria criteria = null;

        criteria =
            createCriteria().createAlias(AllowedTier.DATACENTER_LIMIT_PROPERTY, LIMIT_ALIAS).add(
                Restrictions.eq(LIMIT_ALIAS + "." + PersistentEntity.ID_PROPERTY, limit.getId()));
        criteria.createAlias(AllowedTier.TIER_PROPERTY, "tier");
        criteria.setProjection(Projections.projectionList()
            .add(Projections.property("tier." + PersistentEntity.ID_PROPERTY))
            .add(Projections.property("tier." + Tier.NAME_PROPERTY))
            .add(Projections.property("tier." + Tier.DESCRIPTION_PROPERTY))
            .add(Projections.property("tier." + Tier.DATACENTER_PROPERTY))
            .add(Projections.property("tier." + Tier.ENABLED_PROPERTY))
            .add(Projections.property("tier." + Tier.DEFAULT_ALLOWED_PROPERTY)));
        try
        {
            criteria.setResultTransformer(new AliasToBeanConstructorResultTransformer(Tier.class
                .getConstructor(Integer.class, String.class, String.class, Datacenter.class,
                    boolean.class, boolean.class)));
        }
        catch (NoSuchMethodException ex)
        {
            throw new PersistenceException(ex.getMessage(), ex.getCause());
        }

        return criteria.list();
    }

    /**
     * Returns the enabled tiers by an enterprise an a datacenter. It's important to know that the
     * result object is a list of {@link Tier} class objects. Not {@link AllowedTier} as could be
     * expected.
     * 
     * @param enterpriseId
     * @param datacenterId
     * @return list of {@link Tier}
     */
    @SuppressWarnings("unchecked")
    public List<Tier> getEnabledTiers(final Integer enterpriseId, final Integer datacenterId)
    {
        Criteria criteria = createTiersCriteria(enterpriseId, datacenterId, Boolean.TRUE);
        return criteria.list();
    }

    /**
     * Return the enabled tier by enterprise, datacenter and tier identifier. It's important to know
     * that the result object is a {@link Tier} class objects. Not {@link AllowedTier} as could be
     * expected.
     * 
     * @param enterpriseId
     * @param datacenterId
     * @param tierId
     * @return {@link Tier}
     */
    public Tier findTier(final Integer enterpriseId, final Integer datacenterId,
        final Integer tierId, final boolean onlyEnabled)
    {
        Criteria criteria = createTiersCriteria(enterpriseId, datacenterId, onlyEnabled);
        criteria.add(Restrictions.eq("tier." + PersistentEntity.ID_PROPERTY, tierId));
        return (Tier) criteria.uniqueResult();
    }

    private Criteria createTiersCriteria(final Integer enterpriseId, final Integer datacenterId,
        final boolean onlyEnabled)
    {
        Criteria criteria =
            createCriteria()
                // limit alias
                .createAlias(AllowedTier.DATACENTER_LIMIT_PROPERTY, LIMIT_ALIAS)
                .add(
                    Restrictions.eq(LIMIT_ALIAS + "." + DatacenterLimits.ENTERPRISE_PROPERTY + "."
                        + PersistentEntity.ID_PROPERTY, enterpriseId))
                .add(
                    Restrictions.eq(LIMIT_ALIAS + "." + DatacenterLimits.DATACENTER_PROPERTY + "."
                        + PersistentEntity.ID_PROPERTY, datacenterId))
                // tier alias
                .createAlias(AllowedTier.TIER_PROPERTY, "tier")
                // tier projection
                .setProjection(
                    Projections.projectionList()
                        .add(Projections.property("tier." + PersistentEntity.ID_PROPERTY))
                        .add(Projections.property("tier." + Tier.NAME_PROPERTY))
                        .add(Projections.property("tier." + Tier.DESCRIPTION_PROPERTY))
                        .add(Projections.property("tier." + Tier.DATACENTER_PROPERTY))
                        .add(Projections.property("tier." + Tier.ENABLED_PROPERTY))
                        .add(Projections.property("tier." + Tier.DEFAULT_ALLOWED_PROPERTY)));
        if (onlyEnabled)
        {
            criteria.add(Restrictions.eq("tier." + Tier.ENABLED_PROPERTY, onlyEnabled));
        }
        try
        {
            // return instances of Tier.class
            criteria.setResultTransformer(new AliasToBeanConstructorResultTransformer(Tier.class
                .getConstructor(Integer.class, String.class, String.class, Datacenter.class,
                    boolean.class, boolean.class)));
            return criteria;
        }
        catch (NoSuchMethodException ex)
        {
            throw new PersistenceException(ex.getMessage(), ex.getCause());
        }
    }
}
