/**
 * Copyright (C) 2008 - Abiquo Holdings S.L. All rights reserved.
 *
 * Please see /opt/abiquo/tomcat/webapps/legal/ on Abiquo server
 * or contact contact@abiquo.com for licensing information.
 */
package com.abiquo.server.core.enterprise;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import javax.persistence.EntityManager;

import org.hibernate.Query;
import org.hibernate.criterion.Criterion;
import org.hibernate.criterion.Restrictions;
import org.springframework.stereotype.Repository;

import com.abiquo.server.core.common.persistence.DefaultDAOBase;
import com.abiquo.server.core.infrastructure.Datacenter;
import com.abiquo.server.core.util.FilterOptions;
import com.abiquo.server.core.util.PagedList;
import com.softwarementors.bzngine.entities.PersistentEntity;

@Repository("jpaDatacenterLimitsDAO")
public class DatacenterLimitsDAO extends DefaultDAOBase<Integer, DatacenterLimits>
{
    private static final String ENTERPRISES_BY_TIER =
        "SELECT lim.enterprise FROM com.abiquo.server.core.enterprise.DatacenterLimits lim, "
            + "com.abiquo.server.core.enterprise.AllowedTier tial "
            + "WHERE lim.id = tial.datacenterLimit.id AND tial.tier.id = :idTier AND lim.enterprise.name like :name order by lim.enterprise.name asc";

    private static final String ENTERPRISES_BY_TIER_AND_SCOPE =
        "SELECT lim.enterprise FROM com.abiquo.server.core.enterprise.DatacenterLimits lim, "
            + "com.abiquo.server.core.enterprise.ScopeEntity entity, "
            + "com.abiquo.server.core.enterprise.AllowedTier tial "
            + "WHERE lim.id = tial.datacenterLimit.id AND tial.tier.id = :idTier "
            + "AND lim.enterprise.id = entity.idResource AND entity.type = 'ENTERPRISE' "
            + "AND entity.scope.id = :idScope AND lim.enterprise.name like :name order by lim.enterprise.name asc";

    public DatacenterLimitsDAO()
    {
        super(DatacenterLimits.class);
    }

    public DatacenterLimitsDAO(final EntityManager entityManager)
    {
        super(DatacenterLimits.class, entityManager);
    }

    private Criterion sameEnterprise(final Enterprise enterprise)
    {
        return Restrictions.eq(DatacenterLimits.ENTERPRISE_PROPERTY, enterprise);
    }

    private Criterion sameDatacenter(final Datacenter datacenter)
    {
        return Restrictions.eq(DatacenterLimits.DATACENTER_PROPERTY, datacenter);
    }

    private Criterion sameId(final Integer id)
    {
        return Restrictions.eq(PersistentEntity.ID_PROPERTY, id);
    }

    public Collection<DatacenterLimits> findByDatacenter(final Datacenter datacenter)
    {
        return findByCriterions(sameDatacenter(datacenter));
    }

    public Collection<DatacenterLimits> findByEnterprise(final Enterprise enterprise)
    {
        return findByCriterions(sameEnterprise(enterprise), Restrictions
            .isNotNull(DatacenterLimits.DATACENTER_PROPERTY));
    }

    public DatacenterLimits findByEnterpriseAndDatacenter(final Enterprise enterprise,
        final Datacenter datacenter)
    {
        return (DatacenterLimits) createCriteria(sameEnterprise(enterprise),
            sameDatacenter(datacenter)).uniqueResult();
    }

    public DatacenterLimits findByEnterpriseAndIdentifier(final Enterprise enterprise,
        final Integer limitId)
    {
        return (DatacenterLimits) createCriteria(sameEnterprise(enterprise), sameId(limitId))
            .uniqueResult();
    }

    public Collection<Enterprise> getAllowedEnterprisesByTiers(final Integer idTier,
        final FilterOptions filterOptions, final Integer idScope)
    {
        /*
         * try {
         */
        Query query = null;
        Map<String, Object> params = new HashMap<String, Object>();

        params.put("idTier", idTier);
        params.put("name", filterOptions == null || filterOptions.getFilter() == null ? "%" : "%"
            + filterOptions.getFilter() + "%");
        if (Scope.DEFAULT_SCOPE_ID.equals(idScope))
        {
            query = getSession().createQuery(ENTERPRISES_BY_TIER);
        }
        else
        {
            query = getSession().createQuery(ENTERPRISES_BY_TIER_AND_SCOPE);
            params.put("idScope", idScope);
        }
        query.setProperties(params);

        // FIXME: scastro needs help from zmalik to understand how scopes work
        // Actually the filter works by the sql definition.
        /*
         * if (idScope != 1) { JPAConfiguration.enableScopeFilter(getEntityManager(), idScope); }
         */

        if (filterOptions == null)
        {
            return query.list();
        }

        // Check if the page requested is bigger than the last one
        Integer totalResults = count(query, params).intValue();
        // Limit 0 means no size filter
        if (filterOptions.getLimit() == 0)
        {
            filterOptions.setLimit(totalResults);
            filterOptions.setStartwith(0);
        }
        else if (filterOptions.getStartwith() >= totalResults)
        {
            filterOptions.setStartwith(totalResults - filterOptions.getLimit());
        }

        query.setFirstResult(filterOptions.getStartwith());
        query.setMaxResults(filterOptions.getLimit());

        PagedList<Enterprise> enterpriseList = new PagedList<Enterprise>(query.list());
        enterpriseList.setTotalResults(totalResults);
        enterpriseList.setPageSize(filterOptions.getLimit());
        enterpriseList.setCurrentElement(filterOptions.getStartwith());

        return enterpriseList;
        /*
         * } finally { JPAConfiguration.disableScopeFilter(getEntityManager()); }
         */
    }
}
