/**
 * Copyright (C) 2008 - Abiquo Holdings S.L. All rights reserved.
 *
 * Please see /opt/abiquo/tomcat/webapps/legal/ on Abiquo server
 * or contact contact@abiquo.com for licensing information.
 */
package com.abiquo.server.core.enterprise;

import java.util.Collection;
import java.util.List;

import javax.persistence.EntityManager;

import org.apache.commons.lang.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Repository;

import com.abiquo.model.enumerator.AuthType;
import com.abiquo.server.core.common.DefaultEntityCurrentUsed;
import com.abiquo.server.core.common.DefaultRepBase;
import com.abiquo.server.core.common.persistence.JPAConfiguration;
import com.abiquo.server.core.infrastructure.Datacenter;
import com.abiquo.server.core.infrastructure.Machine;
import com.abiquo.server.core.infrastructure.MachineDAO;
import com.abiquo.server.core.infrastructure.storage.Tier;
import com.abiquo.server.core.pricing.PricingTemplate;

@Repository
public class EnterpriseRep extends DefaultRepBase
{

    /* package: test only */static final String BUG_INSERT_NAME_MUST_BE_UNIQUE =
        "ASSERT- insert: enterprise name must be unique";

    /* package: test only */static final String BUG_UPDATE_NAME_MUST_BE_UNIQUE =
        "ASSERT- update: enterprise name must be unique";

    @Autowired
    private EnterpriseDAO enterpriseDAO;

    @Autowired
    private PrivilegeDAO privilegeDAO;

    @Autowired
    private RoleDAO roleDAO;

    @Autowired
    private RoleLdapDAO roleLdapDAO;

    @Autowired
    private UserDAO userDAO;

    @Autowired
    private MachineDAO machineDAO;

    @Autowired
    private EnterprisePropertiesDAO enterprisePropertiesDAO;

    @Autowired
    private DatacenterLimitsDAO limitsDAO;

    @Autowired
    private OneTimeTokenSessionDAO ottSessionDAO;

    @Autowired
    private SessionDAO sessionDAO;

    @Autowired
    private AllowedTierDAO allowedTierDAO;

    public EnterpriseRep()
    {

    }

    public EnterpriseRep(final EntityManager entityManager)
    {
        assert entityManager != null;
        assert entityManager.isOpen();

        this.entityManager = entityManager;
        this.enterpriseDAO = new EnterpriseDAO(entityManager);
        userDAO = new UserDAO(entityManager);
        roleDAO = new RoleDAO(entityManager);
        ottSessionDAO = new OneTimeTokenSessionDAO(entityManager);
        privilegeDAO = new PrivilegeDAO(entityManager);
        roleLdapDAO = new RoleLdapDAO(entityManager);
        sessionDAO = new SessionDAO(entityManager);
        allowedTierDAO = new AllowedTierDAO(entityManager);
    }

    public void insert(final Enterprise enterprise)
    {
        assert enterprise != null;
        assert !enterpriseDAO.isManaged(enterprise);
        assert !enterpriseDAO.existsAnyWithName(enterprise.getName()) : BUG_INSERT_NAME_MUST_BE_UNIQUE;

        enterpriseDAO.persist(enterprise);
        enterpriseDAO.flush();
    }

    public void update(final Enterprise enterprise)
    {
        assert enterprise != null;
        assert enterpriseDAO.isManaged(enterprise);
        assert !enterpriseDAO.existsAnyOtherWithName(enterprise, enterprise.getName()) : BUG_UPDATE_NAME_MUST_BE_UNIQUE;

        enterpriseDAO.flush();
    }

    public EnterpriseProperties findPropertiesByEnterprise(final Enterprise enterprise)
    {
        return enterprisePropertiesDAO.findByEnterprise(enterprise);
    }

    public void updateEnterpriseProperties(final EnterpriseProperties enterpriseProperties)
    {
        enterprisePropertiesDAO.flush();
    }

    public void removeEnterpriseProperties(final EnterpriseProperties ep)
    {
        enterprisePropertiesDAO.remove(ep);
    }

    public void createEnterpriseProperties(final Enterprise enterprise)
    {
        enterprisePropertiesDAO.persist(new EnterpriseProperties(enterprise));
    }

    public RoleLdap findLdapRoleByType(final String type)
    {
        return roleLdapDAO.findByType(type);
    }

    public Enterprise findById(final Integer id)
    {
        return findById(id, 1); // default
    }

    public Enterprise findById(final Integer id, final int idScope)
    {
        try
        {
            if (!Scope.DEFAULT_SCOPE_ID.equals(idScope))
            {
                JPAConfiguration.enableScopeFilter(entityManager, idScope);
            }
            assert id != null;

            return this.enterpriseDAO.findById(id);
        }
        finally
        {
            JPAConfiguration.disableScopeFilter(entityManager);
        }
    }

    public List<Enterprise> findAll()
    {
        return this.enterpriseDAO.findAll();
    }

    public List<Enterprise> findAll(final Integer startwith, final Integer numResults,
        final int idScope)
    {
        return this.enterpriseDAO.findAll(startwith, numResults, idScope);
    }

    public List<Enterprise> findByPricingTemplate(final Integer firstElem,
        final PricingTemplate pricingTempl, final boolean included, final String filterName,
        final Integer numResults, final Integer idEnterprise, final int idScope)
    {
        return this.enterpriseDAO.findByPricingTemplate(firstElem, pricingTempl, included,
            filterName, numResults, idEnterprise, idScope);
    }

    public List<Enterprise> findByNameAnywhere(final String name, final int idScope)
    {
        assert name != null;

        List<Enterprise> result = this.enterpriseDAO.findByNameAnywhere(name, idScope);
        return result;
    }

    public boolean existsAnyOtherWithName(final Enterprise enterprise, final String name)
    {
        assert enterprise != null;
        assert !StringUtils.isEmpty(name);

        return this.enterpriseDAO.existsAnyOtherWithName(enterprise, name);
    }

    public boolean existsAnyWithName(final String name)
    {
        assert !StringUtils.isEmpty(name);

        return this.enterpriseDAO.existsAnyWithName(name);
    }

    public void delete(final Enterprise enterprise)
    {
        assert enterprise != null;
        assert enterpriseDAO.isManaged(enterprise);

        enterpriseDAO.remove(enterprise);
        enterpriseDAO.flush();
    }

    public Collection<User> findAllUsers()
    {
        return userDAO.findAll();
    }

    public Collection<User> findUsersByRole(final Role role)
    {
        return userDAO.findByRole(role);
    }

    public Collection<User> findUsersByEnterprise(final Enterprise enterprise)
    {
        return userDAO.findByEnterprise(enterprise);
    }

    public Collection<User> findUsersByEnterprise(final Enterprise enterprise, final Role role,
        final String filter, final String order, final boolean desc, final boolean connected,
        final Integer page, final Integer numResults, final Collection<Enterprise> enterprises)
    {
        return userDAO.find(enterprise, role, filter, order, desc, connected, page, numResults,
            enterprises);
    }

    public User findUserByEnterprise(final Integer userId, final Enterprise enterprise)
    {
        return userDAO.findByEnterprise(userId, enterprise);
    }

    public boolean existAnyUserWithNick(final String nick)
    {
        return userDAO.existAnyUserWithNick(nick);
    }

    public boolean existAnyOtherUserWithNick(final User user, final String nick)
    {
        return userDAO.existAnyOtherUserWithNick(user, nick);
    }

    public boolean existAnyUserWithRole(final Role role)
    {
        return userDAO.existAnyUserWithRole(role);
    }

    public void insertUser(final User user)
    {
        userDAO.persist(user);
    }

    public void updateUser(final User user)
    {
        userDAO.flush();
    }

    public void removeUser(final User user)
    {
        userDAO.remove(user);
    }

    public User findUserById(final Integer id)
    {
        return userDAO.findById(id);
    }

    public Role findRoleById(final Integer id)
    {
        return roleDAO.findById(id);
    }

    public Collection<Role> findAllRoles()
    {
        return roleDAO.findAll();
    }

    public Collection<Role> findRolesByEnterprise(final Enterprise enterprise, final String filter,
        final String order, final boolean desc, final Integer page, final Integer numResults)
    {
        return roleDAO.find(enterprise, filter, order, desc, page, numResults);
    }

    public Collection<Role> findRolesByEnterpriseNotNull(final Enterprise enterprise,
        final String filter, final String order, final boolean desc, final Integer page,
        final Integer numResults)
    {
        return roleDAO.find(enterprise, filter, order, desc, page, numResults, true);
    }

    public Collection<Role> findRolesByEnterpriseAndName(final Enterprise enterprise,
        final String filter, final String order, final boolean desc, final Integer page,
        final Integer numResults, final boolean discardNullEnterprises)
    {
        return roleDAO.findExactly(enterprise, filter, order, desc, page, numResults,
            discardNullEnterprises);
    }

    public Collection<Role> findRolesByEnterpriseAndName(final Enterprise enterprise,
        final String filter, final String order, final boolean desc, final Integer page,
        final Integer numResults)
    {
        return findRolesByEnterpriseAndName(enterprise, filter, order, desc, page, numResults, true);
    }

    public void insertRole(final Role role)
    {
        roleDAO.persist(role);
    }

    public void updateRole(final Role role)
    {
        roleDAO.flush();
    }

    public void deleteRole(final Role role)
    {
        roleDAO.remove(role);
    }

    public List<Privilege> findPrivilegesByRole(final Role role)
    {
        return roleDAO.findPrivilegesByIdRole(role.getId());
    }

    public Collection<Privilege> findAllPrivileges()
    {
        return privilegeDAO.findAll();
    }

    public Privilege findPrivilegeById(final Integer id)
    {
        return privilegeDAO.findById(id);
    }

    public Collection<RoleLdap> findRolesLdap(final String filter, final String order,
        final boolean desc, final Integer page, final Integer numResults)
    {
        return roleLdapDAO.find(filter, order, desc, page, numResults);
    }

    public Collection<RoleLdap> findRoleLdapByRole(final Role role)
    {
        return roleLdapDAO.findByRole(role);
    }

    public RoleLdap findRoleLdapById(final Integer id)
    {
        return roleLdapDAO.findById(id);
    }

    public List<RoleLdap> findRoleLdapByRoleLdap(final String roleLdap)
    {
        return roleLdapDAO.findByRoleLdap(roleLdap);
    }

    public void insertRoleLdap(final RoleLdap roleLdap)
    {
        roleLdapDAO.persist(roleLdap);
    }

    public void updateRoleLdap(final RoleLdap roleLdap)
    {
        roleLdapDAO.flush();
    }

    public void deleteRoleLdap(final RoleLdap roleLdap)
    {
        roleLdapDAO.remove(roleLdap);
    }

    public boolean existAnyRoleLdapWithRole(final Role role)
    {
        return roleLdapDAO.existAnyRoleLdapWithRole(role);
    }

    public DefaultEntityCurrentUsed getEnterpriseResourceUsage(final int enterpriseId)
    {
        return enterpriseDAO.getEnterpriseResourceUsage(enterpriseId);
    }

    public Enterprise findByName(final String name)
    {
        return enterpriseDAO.findUniqueByProperty(Enterprise.NAME_PROPERTY, name);
    }

    public User getUserByUserName(final String nick)
    {
        return userDAO.findUniqueByProperty(User.NICK_PROPERTY, nick);
    }

    public List<Machine> findReservedMachines(final Enterprise enterprise)
    {
        return machineDAO.findReservedMachines(enterprise);
    }

    public Machine findReservedMachine(final Enterprise enterprise, final Integer machineId)
    {
        return machineDAO.findReservedMachine(enterprise, machineId);
    }

    public void reserveMachine(final Machine machine, final Enterprise enterprise)
    {
        machineDAO.reserveMachine(machine, enterprise);
    }

    public void releaseMachine(final Machine machine)
    {
        machineDAO.releaseMachine(machine);
    }

    public DatacenterLimits findLimitsByEnterpriseAndDatacenter(final Enterprise enterprise,
        final Datacenter datacenter)
    {
        return limitsDAO.findByEnterpriseAndDatacenter(enterprise, datacenter);
    }

    public DatacenterLimits findLimitsByEnterpriseAndIdentifier(final Enterprise enterprise,
        final Integer limitId)
    {
        return limitsDAO.findByEnterpriseAndIdentifier(enterprise, limitId);
    }

    public Collection<DatacenterLimits> findLimitsByEnterprise(final Enterprise enterprise)
    {
        return limitsDAO.findByEnterprise(enterprise);
    }

    public Collection<DatacenterLimits> findLimitsByDatacenter(final Datacenter datacenter)
    {
        return limitsDAO.findByDatacenter(datacenter);
    }

    public void insertLimit(final DatacenterLimits limit)
    {
        limitsDAO.persist(limit);
    }

    public void updateLimit(final DatacenterLimits limit)
    {
        limitsDAO.flush();
    }

    public void deleteLimit(final DatacenterLimits limit)
    {
        limitsDAO.remove(limit);
    }

    /**
     * Consumes the token. After a successful execution the token will be invalidated.
     * 
     * @param token token.
     * @return boolean true if there was token. False otherwise.
     */
    public boolean existOneTimeToken(final String token)
    {
        return ottSessionDAO.consumeToken(token) > 0;
    }

    /**
     * The uniqueness of users is granted by Login + AuthType.
     * 
     * @param token token to persist.
     */
    public void persistToken(final String token)
    {
        OneTimeTokenSession ottSession = new OneTimeTokenSession(token);
        ottSessionDAO.save(ottSession);
    }

    /**
     * {@see UserDAO#getAbiquoUserByLogin(String)}
     */
    public User getAbiquoUserByUserName(final String nick)
    {
        return userDAO.getAbiquoUserByLogin(nick);
    }

    /**
     * {@see UserDAO#getUserByAuth(String, authType)}
     */
    public User getUserByAuth(final String nick, final AuthType authType)
    {
        return userDAO.getUserByAuth(nick, authType);
    }

    /**
     * The uniqueness of users is granted by Login + AuthType.
     * 
     * @param nick login.
     * @param authType an {@link AuthType} value.
     * @return boolean false if there is no other user with same nick + authType. False otherwise.
     */
    public boolean existAnyUserWithNickAndAuth(final String nick, final AuthType authType)
    {
        return userDAO.existAnyUserWithNickAndAuth(nick, authType);
    }

    public boolean existAnyEnterpriseWithPricingTemplate(final PricingTemplate pricingTemplate)
    {
        return enterpriseDAO.existAnyEnterpriseWithPricingTemplate(pricingTemplate);
    }

    public boolean isUserAllowedToUseVirtualDatacenter(final String username,
        final String authtype, final String[] privileges, final Integer idVdc)
    {
        return userDAO.isUserAllowedToUseVirtualDatacenter(username, authtype, privileges, idVdc)
            && userDAO.isVDCwithinScope(username, idVdc);
    }

    public boolean isUserAllowedToEnterprise(final String username, final String authtype,
        final String[] privileges, final Integer idEnterprise)
    {
        return userDAO.isUserAllowedToEnterprise(username, authtype, privileges, idEnterprise);
    }

    public Collection<Enterprise> findByScope(final int idScope, final Integer startwith,
        final String filterName, final Integer numResults)
    {
        return enterpriseDAO.findByScope(idScope, startwith, filterName, numResults);
    }

    public Collection<Enterprise> findAllEnterprisesIdName(final int idScope,
        final Integer startwith, final String filterName, final Integer numResults)
    {
        return this.enterpriseDAO.findByScope(idScope, startwith, filterName, numResults);
    }

    public List<Session> findAllSessions()
    {
        return sessionDAO.findAll();
    }

    public List<Session> findSessionsByKey(final String key)
    {
        return sessionDAO.findByKey(key);
    }

    public Session findSessionByUserAndKey(final User user, final String key)
    {
        return sessionDAO.findByUserAndKey(user, key);
    }

    public void insertSession(final Session session)
    {
        sessionDAO.persist(session);
    }

    public void deleteSession(final Session session)
    {
        sessionDAO.remove(session);
    }

    public void deleteSessions(final Collection<Session> sessions)
    {
        for (Session session : sessions)
        {
            sessionDAO.remove(session);
        }
    }

    public Collection<Enterprise> findAll(final Integer idScope)
    {

        try
        {
            if (!Scope.DEFAULT_SCOPE_ID.equals(idScope))

            {
                JPAConfiguration.enableScopeFilter(entityManager, idScope);
            }
            return findAll();

        }
        finally
        {
            JPAConfiguration.disableScopeFilter(entityManager);
        }
    }

    public Collection<Role> getRolesByScope(final Scope scope)
    {
        return roleDAO.getRolesByScope(scope);

    }

    public boolean isEnterpiseWithinUserScope(final String userName, final Integer idEnt)
    {
        return enterpriseDAO.isEnterpriseWithinUserScope(userName, idEnt);
    }

    public boolean isUserAllowedToUseDatacenter(final String userName, final Integer idDatacenter)
    {
        // TODO Auto-generated method stub
        return userDAO.isUserAllowedToUseDatacenter(userName, idDatacenter);
    }

    public Collection<Session> findSessionByUser(final User user)
    {
        return sessionDAO.findByUser(user);
    }

    public void insertAllowedTier(final AllowedTier allowedTier)
    {
        allowedTierDAO.persist(allowedTier);
    }

    public void deleteAllowedTier(final AllowedTier allowedTier)
    {
        allowedTierDAO.remove(allowedTier);
    }

    public Collection<AllowedTier> getAllowedTiersByTier(final Tier tier)
    {
        return allowedTierDAO.getAllowedTiersByTier(tier);
    }

    public AllowedTier findAllowedTierByEnterpriseAndTier(final Enterprise enterprise,
        final Tier tier)
    {
        return allowedTierDAO.findAllowedTierByEnterpriseAndTier(enterprise, tier);
    }

    public Collection<Tier> getTiersByLimit(final DatacenterLimits limit) throws SecurityException,
        NoSuchMethodException
    {
        return allowedTierDAO.getTiersByLimit(limit);
    }
}
