/**
 * Copyright (C) 2008 - Abiquo Holdings S.L. All rights reserved.
 *
 * Please see /opt/abiquo/tomcat/webapps/legal/ on Abiquo server
 * or contact contact@abiquo.com for licensing information.
 */
package com.abiquo.server.core.enterprise;

import java.util.ArrayList;
import java.util.List;

import javax.persistence.CascadeType;
import javax.persistence.Column;
import javax.persistence.Entity;
import javax.persistence.FetchType;
import javax.persistence.GeneratedValue;
import javax.persistence.Id;
import javax.persistence.OneToMany;
import javax.persistence.Table;

import org.hibernate.validator.constraints.Length;

import com.abiquo.server.core.common.DefaultEntityBase;
import com.softwarementors.validation.constraints.LeadingOrTrailingWhitespace;
import com.softwarementors.validation.constraints.Required;

@Entity
@Table(name = Scope.TABLE_NAME)
@org.hibernate.annotations.Table(appliesTo = Scope.TABLE_NAME)
public class Scope extends DefaultEntityBase
{
    public static final String TABLE_NAME = "scope";

    public static final String SCOPE_ID = "idScope";

    // DO NOT ACCESS: present due to needs of infrastructure support. *NEVER* call from business
    // code
    protected Scope()
    {
        // Just for JPA support
    }

    private final static String ID_COLUMN = "id";

    @Id
    @GeneratedValue
    @Column(name = ID_COLUMN, nullable = false)
    private Integer id;

    @Override
    public Integer getId()
    {
        return this.id;
    }

    @Deprecated
    public void setId(final Integer id)
    {
        this.id = id;
    }

    public final static String NAME_PROPERTY = "name";

    private final static boolean NAME_REQUIRED = true;

    public final static int NAME_LENGTH_MIN = 1;

    public final static int NAME_LENGTH_MAX = 255;

    private final static boolean NAME_LEADING_OR_TRAILING_WHITESPACES_ALLOWED = false;

    private final static String NAME_COLUMN = "name";

    /** default scope name defined in _kinton_ schema. */
    public final static String DEFAULT_SCOPE_NAME = "Global scope";

    /** default scope id defined in _kinton_ schema. */
    public final static Integer DEFAULT_SCOPE_ID = 1;

    @Column(name = NAME_COLUMN, nullable = !NAME_REQUIRED, length = NAME_LENGTH_MAX)
    private String name;

    @Required(value = NAME_REQUIRED)
    @Length(min = NAME_LENGTH_MIN, max = NAME_LENGTH_MAX)
    @LeadingOrTrailingWhitespace(allowed = NAME_LEADING_OR_TRAILING_WHITESPACES_ALLOWED)
    public String getName()
    {
        return this.name;
    }

    public void setName(final String name)
    {
        this.name = name;
    }

    public final static String SCOPE_RESOURCES_PROPERTY = "scopeResources";

    @OneToMany(cascade = CascadeType.ALL, fetch = FetchType.LAZY, orphanRemoval = true, mappedBy = "scope")
    private List<ScopeEntity> scopeResources;

    public void setScopeResources(final List<ScopeEntity> scopeResources)
    {
        if (this.scopeResources == null)
        {
            this.scopeResources = new ArrayList<ScopeEntity>();
        }
        this.scopeResources.clear();
        this.scopeResources.addAll(scopeResources);
    }

    public List<ScopeEntity> getScopeResources()
    {
        return scopeResources;
    }

    public void addScopeResource(final ScopeEntity scopeResource)
    {
        if (scopeResources == null)
        {
            scopeResources = new ArrayList<ScopeEntity>();
        }
        if (!scopeResources.contains(scopeResource))
        {
            scopeResources.add(scopeResource);
        }
    }

    public final static String AUTO_ENT_PROPERTY = "automaticAddEnterprise";

    public final static String AUTO_ENT_COLUMN = "autoEnt";

    private final static boolean AUTO_ENT_REQUIRED = true;

    @Column(name = AUTO_ENT_COLUMN, nullable = false)
    private boolean automaticAddEnterprise = false;

    @Required(value = AUTO_ENT_REQUIRED)
    public boolean isAutomaticaEnterpriseAddEnabled()
    {
        return automaticAddEnterprise;
    }

    public void setautomaticAddEnterprise(final boolean automaticAddEnterprise)
    {
        this.automaticAddEnterprise = automaticAddEnterprise;
    }

    public final static String AUTO_DAT_PROPERTY = "automaticAddDatacenter";

    public final static String AUTO_DAT_COLUMN = "autoDat";

    private final static boolean AUTO_DAT_REQUIRED = true;

    @Column(name = AUTO_DAT_COLUMN, nullable = false)
    private boolean automaticAddDatacenter = false;

    @Required(value = AUTO_DAT_REQUIRED)
    public boolean isAutomaticaDatacenterAddEnabled()
    {
        return automaticAddDatacenter;
    }

    public void setautomaticAddDatacenter(final boolean automaticAddDatacenter)
    {
        this.automaticAddDatacenter = automaticAddDatacenter;
    }

    public static enum OrderByEnum
    {
        NAME;

        public static OrderByEnum fromValue(final String orderBy)
        {
            for (OrderByEnum currentOrder : OrderByEnum.values())
            {
                if (currentOrder.name().equalsIgnoreCase(orderBy))
                {
                    return currentOrder;
                }
            }

            return null;
        }
    }

    /**
     * @return true if the scope is the Global scope with access to all entities
     */
    public Boolean isGlobalScope()
    {
        /**
         * * In kinton a default global scope is added with "<column name="name" value="Global
         * scope" />" So we know if a scope is global scope by its name (name is unique in scope and
         * it is not possible to modify global scope) Using just idScope is not viable in test
         * environment because, in tests the scope is created and deleted for every test. Making the
         * id of scope vary depending on every test. So just for support in test the check by name
         * is added.
         */
        return DEFAULT_SCOPE_NAME.equals(getName()) || DEFAULT_SCOPE_ID.equals(getId());

    }

    // *************************** Mandatory constructors ***********************
    public Scope(final String name)
    {
        setName(name);
        setautomaticAddEnterprise(false);
        setautomaticAddDatacenter(false);
    }

    public Scope(final String name, final boolean automaticAddEnterprise,
        final boolean automaticAddDatacenter)
    {
        setName(name);
        setautomaticAddEnterprise(automaticAddEnterprise);
        setautomaticAddDatacenter(automaticAddDatacenter);
    }

}
