/**
 * Copyright (C) 2008 - Abiquo Holdings S.L. All rights reserved.
 *
 * Please see /opt/abiquo/tomcat/webapps/legal/ on Abiquo server
 * or contact contact@abiquo.com for licensing information.
 */
package com.abiquo.server.core.event;

import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map.Entry;

import javax.persistence.EntityManager;

import org.hibernate.Criteria;
import org.hibernate.criterion.CriteriaSpecification;
import org.hibernate.criterion.MatchMode;
import org.hibernate.criterion.Order;
import org.hibernate.criterion.Restrictions;
import org.springframework.stereotype.Repository;

import com.abiquo.model.enumerator.SeverityType;
import com.abiquo.server.core.common.persistence.DefaultDAOBase;
import com.abiquo.server.core.enterprise.User;
import com.abiquo.server.core.util.FilterOptions;
import com.abiquo.server.core.util.PagedList;

@Repository("jpaEventDAO")
public class EventDAO extends DefaultDAOBase<Integer, Event>
{
    public EventDAO()
    {
        super(Event.class);
    }

    public EventDAO(final EntityManager entityManager)
    {
        super(Event.class, entityManager);
    }

    @SuppressWarnings("unchecked")
    public List<Event> getEvents(final User currentUser, final FilterOptions filterOptions,
        final HashMap<String, String> filters, final Collection<User> listOfUsers,
        final Integer idEnterprise, final Integer idUser,
        final Collection<Integer> listOfEnterprises)
    {
        Criteria crit = createCriteria();

        // Add BETWEEN clauses for dates
        Timestamp dateFrom = calculateDateFrom(filters, currentUser);
        Timestamp dateTo = calculateDateTo(filters);
        crit.add(Restrictions.between(Event.TIMESTAMP_PROPERTY, dateFrom, dateTo));

        // Remove the dates filters from the Map to avoid adding them again later
        filters.remove("datefrom");
        filters.remove("dateto");

        // Add WHERE clauses for LIKE matches
        for (Entry<String, String> filter : filters.entrySet())
        {
            if (filter.getKey().equals(Event.SEVERITY_PROPERTY))
            {
                for (SeverityType s : SeverityType.values())
                {
                    if (s.name().equals(filter.getValue()))
                    {
                        crit.add(Restrictions.eq(filter.getKey(),
                            SeverityType.valueOf(filter.getValue())));
                    }
                }
            }
            else
            {
                crit.add(Restrictions.like(filter.getKey(), filter.getValue(), MatchMode.ANYWHERE));
            }
        }

        // Add WHERE clauses for EQUALS matches
        if (idUser != null)
        {
            crit.add(Restrictions.eq(Event.ID_USER_PROPERTY, idUser));
        }

        if (listOfEnterprises != null && !listOfEnterprises.isEmpty())
        {
            crit.add(Restrictions.in(Event.ID_ENTERPRISE_PROPERTY, listOfEnterprises));
        }

        // Add IN clause to select only users in performedBy filter
        if (listOfUsers != null && !listOfUsers.isEmpty())
        {
            ArrayList<String> listUsersName = new ArrayList<String>();
            for (User user : listOfUsers)
            {
                listUsersName.add(user.getNick());
            }
            crit.add(Restrictions.in(Event.PERFORMEDBY_PROPERTY, listUsersName));
        }

        // Add ORDERBY clause
        crit.addOrder(filterOptions.getAsc() ? Order.asc(filterOptions.getOrderBy()) : Order
            .desc(filterOptions.getOrderBy()));

        Integer size = count(crit).intValue();
        // Restore the criteria
        crit.setProjection(null);
        crit.setResultTransformer(CriteriaSpecification.ROOT_ENTITY);

        crit.setFirstResult(filterOptions.getStartwith());
        crit.setMaxResults(filterOptions.getLimit());

        PagedList<Event> eventsList = new PagedList<Event>(crit.list());
        eventsList.setTotalResults(size);
        eventsList.setPageSize(filterOptions.getLimit() > size ? size : filterOptions.getLimit());
        eventsList.setCurrentElement(filterOptions.getStartwith());

        return eventsList;
    }

    private Timestamp convertDate(final String date, final Integer sec)
    {
        return new Timestamp((Long.valueOf(date) + sec) * 1000);
    }

    /**
     * Return the Timestamp of the begining date. By default, this date is the user's creation date,
     * but if a 'datefrom' filter exists, it must be taken in account only if dafefrom is later than
     * user's creation date.
     * 
     * @param filters The query filters
     * @param currentUser The currentUser
     * @return A Timestamp of the dateFrom to filter
     */
    private Timestamp calculateDateFrom(final HashMap<String, String> filters,
        final User currentUser)
    {
        Timestamp dateFrom = new Timestamp(currentUser.getCreationDate().getTime());
        if (filters.containsKey("datefrom") && !"%".equals(filters.get("datefrom")))
        {
            if (convertDate(filters.get("datefrom"), 0).getTime() > dateFrom.getTime())
            {
                dateFrom = convertDate(filters.get("datefrom"), 0);
            }
        }

        return dateFrom;
    }

    /**
     * Return the Timestamp of the ending date. By default, this date is the current date, but if a
     * 'dateto' filter exists, it must be taken in account.
     * 
     * @param filters The query filters
     * @return A Timestamp of the dateTo to filter
     */
    private Timestamp calculateDateTo(final HashMap<String, String> filters)
    {
        return filters.containsKey("dateto") && !"%".equals(filters.get("dateto")) ? convertDate(
            filters.get("dateto"), 86400) : new Timestamp(new Date().getTime());
    }
}
