/**
 * Copyright (C) 2008 - Abiquo Holdings S.L. All rights reserved.
 *
 * Please see /opt/abiquo/tomcat/webapps/legal/ on Abiquo server
 * or contact contact@abiquo.com for licensing information.
 */

package com.abiquo.server.core.infrastructure;

import javax.persistence.CascadeType;
import javax.persistence.Column;
import javax.persistence.Entity;
import javax.persistence.GeneratedValue;
import javax.persistence.Id;
import javax.persistence.OneToOne;
import javax.persistence.Table;
import javax.persistence.Transient;

import org.hibernate.validator.constraints.Length;
import org.hibernate.validator.constraints.Range;

import com.abiquo.server.core.common.DefaultEntityBase;
import com.softwarementors.validation.constraints.LeadingOrTrailingWhitespace;
import com.softwarementors.validation.constraints.Required;

@Entity
@Table(name = Datastore.TABLE_NAME)
@org.hibernate.annotations.Table(appliesTo = Datastore.TABLE_NAME)
public class Datastore extends DefaultEntityBase
{
    public static final String TABLE_NAME = "datastore";

    public Datastore()
    {
        // never call this!
    }

    private final static String ID_COLUMN = "idDatastore";

    @Id
    @GeneratedValue
    @Column(name = ID_COLUMN, nullable = false)
    private Integer id;

    @Override
    public Integer getId()
    {
        return this.id;
    }

    public final static String NAME_PROPERTY = "name";

    private final static boolean NAME_REQUIRED = false;

    private final static int NAME_LENGTH_MIN = 0;

    private final static int NAME_LENGTH_MAX = 255;

    private final static boolean NAME_LEADING_OR_TRAILING_WHITESPACES_ALLOWED = false;

    private final static String NAME_COLUMN = "name";

    @Column(name = NAME_COLUMN, nullable = !NAME_REQUIRED, length = NAME_LENGTH_MAX)
    private String name;

    @Required(value = NAME_REQUIRED)
    @Length(min = NAME_LENGTH_MIN, max = NAME_LENGTH_MAX)
    @LeadingOrTrailingWhitespace(allowed = NAME_LEADING_OR_TRAILING_WHITESPACES_ALLOWED)
    public String getName()
    {
        return this.name;
    }

    public void setName(final String name)
    {
        this.name = name;
    }

    public final static String DIRECTORY_PROPERTY = "directory";

    private final static boolean DIRECTORY_REQUIRED = false;

    private final static int DIRECTORY_LENGTH_MIN = 0;

    private final static int DIRECTORY_LENGTH_MAX = 255;

    private final static boolean DIRECTORY_LEADING_OR_TRAILING_WHITESPACES_ALLOWED = false;

    private final static String DIRECTORY_COLUMN = "directory";

    @Column(name = DIRECTORY_COLUMN, nullable = !DIRECTORY_REQUIRED, length = DIRECTORY_LENGTH_MAX)
    private String directory;

    @Required(value = DIRECTORY_REQUIRED)
    @Length(min = DIRECTORY_LENGTH_MIN, max = DIRECTORY_LENGTH_MAX)
    @LeadingOrTrailingWhitespace(allowed = DIRECTORY_LEADING_OR_TRAILING_WHITESPACES_ALLOWED)
    public String getDirectory()
    {
        return this.directory;
    }

    public void setDirectory(final String directory)
    {
        this.directory = directory;
    }

    public final static String ROOT_PATH_PROPERTY = "rootPath";

    private final static boolean ROOT_PATH_REQUIRED = true;

    private final static int ROOT_PATH_LENGTH_MIN = 1;

    public final static int ROOT_PATH_LENGTH_MAX = 42;

    private final static boolean ROOT_PATH_LEADING_OR_TRAILING_WHITESPACES_ALLOWED = false;

    private final static String ROOT_PATH_COLUMN = "rootPath";

    @Column(name = ROOT_PATH_COLUMN, nullable = !ROOT_PATH_REQUIRED, length = ROOT_PATH_LENGTH_MAX)
    private String rootPath;

    @Required(value = ROOT_PATH_REQUIRED)
    @Length(min = ROOT_PATH_LENGTH_MIN, max = ROOT_PATH_LENGTH_MAX)
    @LeadingOrTrailingWhitespace(allowed = ROOT_PATH_LEADING_OR_TRAILING_WHITESPACES_ALLOWED)
    public String getRootPath()
    {
        return this.rootPath;
    }

    public void setRootPath(final String rootPath)
    {
        this.rootPath = rootPath;
    }

    public final static String SIZE_PROPERTY = "size";

    /* package */final static String SIZE_COLUMN = "Size";

    /* package */final static long SIZE_MIN = Long.MIN_VALUE;

    /* package */final static long SIZE_MAX = Long.MAX_VALUE;

    /* package */final static boolean SIZE_REQUIRED = false;

    @Column(name = SIZE_COLUMN, nullable = true)
    @Range(min = SIZE_MIN, max = SIZE_MAX)
    private long size;

    @Required(value = SIZE_REQUIRED)
    public long getSize()
    {
        return this.size;
    }

    public void setSize(final long size)
    {
        this.size = size;
    }

    public final static String ENABLED_PROPERTY = "enabled";

    private final static String ENABLED_COLUMN = "enabled";

    private final static boolean ENABLED_REQUIRED = true;

    @Column(name = ENABLED_COLUMN, nullable = false)
    private boolean enabled = true;

    @Required(value = ENABLED_REQUIRED)
    public boolean isEnabled()
    {
        return this.enabled;
    }

    public void setEnabled(final boolean enabled)
    {
        this.enabled = enabled;
    }

    public final static String DATASTORE_UUID_PROPERTY = "datastoreUUID";

    private final static boolean DATASTORE_UUID_REQUIRED = false;

    private final static int DATASTORE_UUID_LENGTH_MIN = 0;

    private final static int DATASTORE_UUID_LENGTH_MAX = 255;

    private final static boolean DATASTORE_UUID_LEADING_OR_TRAILING_WHITESPACES_ALLOWED = false;

    private final static String DATASTORE_UUID_COLUMN = "datastoreUuid";

    @Column(name = DATASTORE_UUID_COLUMN, nullable = !DATASTORE_UUID_REQUIRED, length = DATASTORE_UUID_LENGTH_MAX)
    private String datastoreUUID;

    @Required(value = DATASTORE_UUID_REQUIRED)
    @Length(min = DATASTORE_UUID_LENGTH_MIN, max = DATASTORE_UUID_LENGTH_MAX)
    @LeadingOrTrailingWhitespace(allowed = DATASTORE_UUID_LEADING_OR_TRAILING_WHITESPACES_ALLOWED)
    public String getDatastoreUUID()
    {
        return this.datastoreUUID;
    }

    public void setDatastoreUUID(final String datastoreUUID)
    {
        this.datastoreUUID = datastoreUUID;
    }

    public Datastore(final Machine machine, final String name, final String rootPath,
        final String directory)
    {
        setRootPath(rootPath);
        setName(name);
        setDirectory(directory);
        setMachine(machine);
    }

    @OneToOne(mappedBy = "datastore", cascade = CascadeType.REMOVE)
    private ScheduledResources scheduledResources;

    /**
     * Two datastores are the same if their rootPath are the same as well.
     * 
     * @param ds1 The first datastore to compare with the second one
     * @param ds2 The second datastore
     * @return <code>true</code> if the two datastores' rootPath are equals, <code>false</code>
     *         otherwise
     */
    public static boolean isSameDatastore(final Datastore ds1, final Datastore ds2)
    {
        return ds1.getRootPath() != null && ds1.getRootPath().equals(ds2.getRootPath());
    }

    public static final String MACHINE_PROPERTY = "machine";

    @Transient
    private Machine machine;

    public ScheduledResources getScheduledResources()
    {
        return scheduledResources;
    }

    public void setScheduledResources(final ScheduledResources scheduledResources)
    {
        this.scheduledResources = scheduledResources;
    }

    public Machine getMachine()
    {
        return machine = this.getScheduledResources().getMachine();
    }

    public void setMachine(final Machine machine)
    {
        new ScheduledResources(machine, this);
        this.machine = machine;
    }

    public long getUsedSize()
    {
        return scheduledResources == null ? 0 : scheduledResources.getDsUsed();
    }
    
    /*
     * Should be used only for creating Datastores
     */
    public void setDsUsed(Long dsUsed){
        this.scheduledResources.setDsUsed(dsUsed);
    }
}
