/**
 * Copyright (C) 2008 - Abiquo Holdings S.L. All rights reserved.
 *
 * Please see /opt/abiquo/tomcat/webapps/legal/ on Abiquo server
 * or contact contact@abiquo.com for licensing information.
 */
package com.abiquo.server.core.infrastructure;

import java.util.List;

import javax.persistence.EntityManager;
import javax.persistence.PersistenceException;

import org.apache.commons.lang.StringUtils;
import org.hibernate.Criteria;
import org.hibernate.criterion.Criterion;
import org.hibernate.criterion.Restrictions;
import org.springframework.stereotype.Repository;

import com.abiquo.server.core.common.persistence.DefaultDAOBase;
import com.softwarementors.bzngine.entities.PersistentEntity;

@Repository("jpaDatastoreDAO")
public class DatastoreDAO extends DefaultDAOBase<Integer, Datastore>
{
    public DatastoreDAO()
    {
        super(Datastore.class);
    }

    public DatastoreDAO(final EntityManager entityManager)
    {
        super(Datastore.class, entityManager);
    }

    public List<Datastore> findMachineDatastores(final Machine machine)
    {
        return machine.getDatastores();
    }

    /**
     * Use the ''datastoreUUID'' property in order to return all the representations of a shared
     * datastore.
     * 
     * @return all the datastores with the same ''datastore uuid'' INCLUDING the provided datastore.
     */
    public List<Datastore> findShares(final Datastore datastore)
    {
        if (datastore.getDatastoreUUID() == null)
        {
            throw new PersistenceException("Datastore doesn't have set the UUID");
        }

        Criteria crit = createCriteria(sharedUuid(datastore.getDatastoreUUID()));

        return getResultList(crit);
    }

    /***
     * Get the datastore with the provided UUID mounted on the target machine. (this call is
     * expected after ''datastoreSelection'' so the target machine will have the datastore)
     */
    public Datastore findDatastore(final String uuid, final Machine machine)
    {

        Criteria criteria = createCriteria(sharedUuid(uuid));
        List<Datastore> sharedDSs = getResultList(criteria);
        for (Datastore ds : sharedDSs)
        {
            if (ds.getMachine().getId().equals(machine.getId()))
            {
                return ds;
            }
        }
        return null;
    }

    private static Criterion sharedUuid(final String uuid)
    {
        return Restrictions.eq(Datastore.DATASTORE_UUID_PROPERTY, uuid);
    }

    private static Criterion equalName(final String name)
    {
        assert !StringUtils.isEmpty(name);

        return Restrictions.eq(Datastore.NAME_PROPERTY, name);
    }

    public boolean existsAnyWithName(final String name)
    {
        assert !StringUtils.isEmpty(name);

        return this.existsAnyByCriterions(equalName(name));
    }

    public boolean existsAnyOtherWithName(final Datastore entity, final String name)
    {
        assert entity != null;
        assert isManaged(entity);
        assert !StringUtils.isEmpty(name);

        return this.existsAnyOtherByCriterions(entity, equalName(name));
    }

    private static Criterion equalDirectory(final String directory)
    {
        return Restrictions.eq(Datastore.DIRECTORY_PROPERTY, directory);
    }

    public boolean existsAnyWithDirectory(final String directory)
    {
        assert !StringUtils.isEmpty(directory);

        return this.existsAnyByCriterions(equalDirectory(directory));
    }

    public boolean existsAnyOtherWithDirectory(final Datastore entity, final String directory)
    {
        assert entity != null;
        assert isManaged(entity);
        assert !StringUtils.isEmpty(directory);

        return this.existsAnyOtherByCriterions(entity, equalDirectory(directory));
    }

    private static Criterion equalRootPath(final String rootPath)
    {
        assert !StringUtils.isEmpty(rootPath);

        return Restrictions.eq(Datastore.ROOT_PATH_PROPERTY, rootPath);
    }

    public boolean existsAnyWithRootPath(final String rootPath, final Integer machineId)
    {
        assert !StringUtils.isEmpty(rootPath);
        assert machineId != null;

        Criteria cri = createCriteria().add(equalRootPath(rootPath));

        for (Object o : cri.list())
        {
            Datastore d = (Datastore) o;
            if (d.getMachine().getId().equals(machineId))
            {
                return true;
            }
        }
        return false;
    }

    public boolean existsAnyOtherWithRootPath(final Datastore entity, final String rootPath,
        final Integer machineId)
    {
        assert entity != null;
        assert isManaged(entity);
        assert !StringUtils.isEmpty(rootPath);
        assert machineId != null;

        Criteria cri =
            createCriteria().add(equalRootPath(rootPath)).add(
                Restrictions.ne(PersistentEntity.ID_PROPERTY, entity.getId()));

        for (Object o : cri.list())
        {
            Datastore d = (Datastore) o;
            if (d.getMachine().getId().equals(machineId))
            {
                return true;
            }
        }
        return false;
    }

    @Override
    public void persist(final Datastore entity)
    {
        // This asserts are needed for JpaDaoTestBase automated tests
        assert entity != null; // java.lang.IllegalArgumentException: attempt to create create event
                               // with null entity
        assert isInReadWriteTransaction(); // org.hibernate.HibernateException: Illegal attempt to
                                           // associate a collection with two open sessions

        getEntityManager().persist(entity);

        Machine machine = entity.getScheduledResources().getMachine();
        if (machine.getId() == null)
        {
            getEntityManager().persist(machine);
        }

        entity.getScheduledResources().setDatastore(entity);
        entity.getScheduledResources().setMachine(machine);
        getEntityManager().persist(entity.getScheduledResources());

    }

}
