/**
 * Copyright (C) 2008 - Abiquo Holdings S.L. All rights reserved.
 *
 * Please see /opt/abiquo/tomcat/webapps/legal/ on Abiquo server
 * or contact contact@abiquo.com for licensing information.
 */
package com.abiquo.server.core.infrastructure.network;

import java.util.ArrayList;
import java.util.List;

import javax.persistence.CascadeType;
import javax.persistence.Column;
import javax.persistence.Entity;
import javax.persistence.Enumerated;
import javax.persistence.FetchType;
import javax.persistence.GeneratedValue;
import javax.persistence.Id;
import javax.persistence.JoinColumn;
import javax.persistence.JoinTable;
import javax.persistence.ManyToMany;
import javax.persistence.ManyToOne;
import javax.persistence.OneToMany;
import javax.persistence.Table;

import org.apache.commons.lang.builder.ToStringBuilder;
import org.apache.commons.lang.builder.ToStringStyle;
import org.hibernate.annotations.ForeignKey;
import org.hibernate.annotations.Formula;
import org.hibernate.validator.constraints.Length;

import com.abiquo.model.enumerator.NetworkType;
import com.abiquo.server.core.common.DefaultEntityBase;
import com.abiquo.server.core.enterprise.Enterprise;
import com.abiquo.server.core.infrastructure.Datacenter;
import com.softwarementors.validation.constraints.LeadingOrTrailingWhitespace;
import com.softwarementors.validation.constraints.Required;

@Entity
@Table(name = VLANNetwork.TABLE_NAME, uniqueConstraints = {})
// TODO: specify unique constraints
@org.hibernate.annotations.Table(appliesTo = VLANNetwork.TABLE_NAME, indexes = {})
// TODO: specify indexes
public class VLANNetwork extends DefaultEntityBase
{

    // ****************************** JPA support *******************************
    public static final String TABLE_NAME = "vlan_network";

    // DO NOT ACCESS: present due to needs of infrastructure support. *NEVER* call from business
    // code
    public VLANNetwork()
    {
        // Just for JPA support
    }

    private final static String ID_COLUMN = "vlan_network_id";

    @Id
    @GeneratedValue
    @Column(name = ID_COLUMN, nullable = false)
    private Integer id;

    @Override
    public Integer getId()
    {
        return this.id;
    }

    public void setId(final Integer id)
    {
        this.id = id;
    }

    // ******************************* Properties *******************************
    public final static String NAME_PROPERTY = "name";

    private final static boolean NAME_REQUIRED = true;

    private final static int NAME_LENGTH_MIN = 1;

    private final static int NAME_LENGTH_MAX = 40;

    private final static boolean NAME_LEADING_OR_TRAILING_WHITESPACES_ALLOWED = false;

    private final static String NAME_COLUMN = "network_name";

    @Column(name = NAME_COLUMN, nullable = !NAME_REQUIRED, length = NAME_LENGTH_MAX)
    private String name;

    @Required(value = NAME_REQUIRED)
    @Length(min = NAME_LENGTH_MIN, max = NAME_LENGTH_MAX)
    @LeadingOrTrailingWhitespace(allowed = NAME_LEADING_OR_TRAILING_WHITESPACES_ALLOWED)
    public String getName()
    {
        return this.name;
    }

    public void setName(final String name)
    {
        this.name = name;
    }

    public final static String TAG_PROPERTY = "tag";

    private final static boolean TAG_REQUIRED = false;

    private final static String TAG_COLUMN = "vlan_tag";

    public static final Integer VLAN_MAX_TAG = 4094;

    @Column(name = TAG_COLUMN, nullable = !TAG_REQUIRED)
    private Integer tag;

    public Integer getTag()
    {
        return this.tag;
    }

    public void setTag(final Integer tag)
    {
        this.tag = tag;
    }

    public final static String TYPE_PROPERTY = "type";

    private final static boolean TYPE_REQUIRED = true;

    private final static String TYPE_COLUMN = "networktype";

    private final static int TYPE_COLUMN_LENGTH = 255;

    @Enumerated(value = javax.persistence.EnumType.STRING)
    @Column(name = TYPE_COLUMN, nullable = !TYPE_REQUIRED, length = TYPE_COLUMN_LENGTH)
    private NetworkType type;

    @Required(value = TYPE_REQUIRED)
    public NetworkType getType()
    {
        return this.type;
    }

    public void setType(final NetworkType type)
    {
        this.type = type;
    }

    // ****************************** Associations ******************************
    public final static String NETWORK_PROPERTY = "network";

    private final static boolean NETWORK_REQUIRED = true;

    private final static String NETWORK_ID_COLUMN = "network_id";

    @JoinColumn(name = NETWORK_ID_COLUMN)
    @ManyToOne(fetch = FetchType.LAZY)
    @ForeignKey(name = "FK_" + TABLE_NAME + "_network")
    private Network network;

    @Required(value = NETWORK_REQUIRED)
    public Network getNetwork()
    {
        return this.network;
    }

    public void setNetwork(final Network network)
    {
        this.network = network;
    }

    // `enterprise_id` int(10) unsigned DEFAULT NULL,

    public final static String ENTERPRISE_PROPERTY = "enterprise";

    private final static boolean ENTERPRISE_REQUIRED = false;

    private final static String ENTERPRISE_ID_COLUMN = "enterprise_id";

    @JoinColumn(name = ENTERPRISE_ID_COLUMN)
    @ManyToOne(fetch = FetchType.LAZY)
    @ForeignKey(name = "FK_" + TABLE_NAME + "_enterprise")
    private Enterprise enterprise;

    @Required(value = ENTERPRISE_REQUIRED)
    public Enterprise getEnterprise()
    {
        return this.enterprise;
    }

    public void setEnterprise(final Enterprise enterprise)
    {
        this.enterprise = enterprise;
    }

    //

    public final static String CONFIGURATION_PROPERTY = "configuration";

    private final static boolean CONFIGURATION_REQUIRED = true;

    private final static String CONFIGURATION_ID_COLUMN = "network_configuration_id";

    @JoinColumn(name = CONFIGURATION_ID_COLUMN)
    @ManyToOne(fetch = FetchType.LAZY, cascade = CascadeType.ALL)
    @ForeignKey(name = "FK_" + TABLE_NAME + "_configuration")
    private NetworkConfiguration configuration;

    @Required(value = CONFIGURATION_REQUIRED)
    public NetworkConfiguration getConfiguration()
    {
        return this.configuration;
    }

    public void setConfiguration(final NetworkConfiguration configuration)
    {
        this.configuration = configuration;
    }

    private static final String LIMIT_ID_QUERY =
        "select limits.idDatacenterLimit from enterprise_limits_by_datacenter limits inner join datacenter on limits.idDataCenter = datacenter.idDataCenter inner join enterprise on limits.idEnterprise = enterprise.idEnterprise inner join network on datacenter.network_id = network.network_id inner join vlan_network vlan on network.network_id = vlan.network_id where vlan.vlan_network_id = vlan_network_id and vlan.enterprise_id = enterprise.idEnterprise";

    /**
     * External networks are logically assigned to an {@link Enterprise} and a {@link Datacenter}.
     * The limit id identifies the limit entity that links both entities with this network.
     */
    @Formula("(" + LIMIT_ID_QUERY + ")")
    private Integer limitId;

    // This should never be set. It is a readonly value computed by the formula
    private void setLimitId(final Integer limitId)
    {
        this.limitId = limitId;
    }

    public Integer getLimitId()
    {
        return limitId;
    }

    // I don't want to access the ips directly but I want to remove them in cascade
    @OneToMany(cascade = CascadeType.ALL, fetch = FetchType.LAZY, orphanRemoval = true, mappedBy = "vlanNetwork")
    private List<IpPoolManagement> ipPoolManagements;

    public final static String ASSOCIATION_TABLE = "vlans_dhcpOption";

    @ManyToMany(fetch = FetchType.LAZY, targetEntity = DhcpOption.class, cascade = CascadeType.DETACH)
    @JoinTable(name = ASSOCIATION_TABLE, joinColumns = @JoinColumn(name = "idVlan"), inverseJoinColumns = @JoinColumn(name = "idDhcpOption"))
    private List<DhcpOption> dhcpOptions = new ArrayList<DhcpOption>();

    public List<DhcpOption> getDhcpOption()
    {
        return dhcpOptions;
    }

    public void setDhcpOption(final List<DhcpOption> dhcpOptions)
    {
        this.dhcpOptions = dhcpOptions;
    }

    public void addDhcpOption(final DhcpOption dhcpOption)
    {
        if (dhcpOptions == null)
        {
            dhcpOptions = new ArrayList<DhcpOption>();
        }
        dhcpOptions.add(dhcpOption);
    }

    public final static String NETWORK_SERVICE_TYPE_PROPERTY = "networkServiceType";

    private final static boolean NETWORK_SERVICE_TYPE_REQUIRED = false;

    private final static String NETWORK_SERVICE_TYPE_ID_COLUMN = "network_service_type_id";

    @JoinColumn(name = NETWORK_SERVICE_TYPE_ID_COLUMN)
    @ManyToOne(fetch = FetchType.LAZY)
    @ForeignKey(name = "vlannetwork_network_service_type_FK")
    private NetworkServiceType networkServiceType;

    @Required(value = NETWORK_SERVICE_TYPE_REQUIRED)
    public NetworkServiceType getNetworkServiceType()
    {
        return this.networkServiceType;
    }

    public void setNetworkServiceType(final NetworkServiceType networkServiceType)
    {
        this.networkServiceType = networkServiceType;
    }

    // ********************************** Others ********************************
    @Override
    public String toString()
    {
        return ToStringBuilder.reflectionToString(this, ToStringStyle.MULTI_LINE_STYLE);
    }

    // *************************** Mandatory constructors ***********************
    public VLANNetwork(final String name, final Network network, final NetworkType type,
        final NetworkConfiguration configuration, final NetworkServiceType serviceType)
    {
        setName(name);
        setTag(tag);
        setNetwork(network);
        setType(type);
        setConfiguration(configuration);
        setNetworkServiceType(serviceType);
    }
}
