/**
 * Copyright (C) 2008 - Abiquo Holdings S.L. All rights reserved.
 *
 * Please see /opt/abiquo/tomcat/webapps/legal/ on Abiquo server
 * or contact contact@abiquo.com for licensing information.
 */
package com.abiquo.server.core.infrastructure.storage;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.persistence.EntityManager;

import org.hibernate.Criteria;
import org.hibernate.Query;
import org.hibernate.SQLQuery;
import org.hibernate.Session;
import org.hibernate.criterion.Criterion;
import org.hibernate.criterion.Restrictions;
import org.springframework.stereotype.Repository;

import com.abiquo.model.enumerator.VolumeState;
import com.abiquo.server.core.cloud.VirtualAppliance;
import com.abiquo.server.core.cloud.VirtualDatacenter;
import com.abiquo.server.core.cloud.VirtualMachine;
import com.abiquo.server.core.common.persistence.DefaultDAOBase;
import com.abiquo.server.core.enterprise.Enterprise;
import com.abiquo.server.core.enterprise.User;
import com.abiquo.server.core.infrastructure.management.Rasd;
import com.abiquo.server.core.infrastructure.management.RasdManagement;
import com.abiquo.server.core.util.FilterOptions;
import com.abiquo.server.core.util.PagedList;
import com.softwarementors.bzngine.entities.PersistentEntity;

@Repository("jpaVolumeManagementDAO")
/* package */class VolumeManagementDAO extends DefaultDAOBase<Integer, VolumeManagement>
{
    public VolumeManagementDAO()
    {
        super(VolumeManagement.class);
    }

    public VolumeManagementDAO(final EntityManager entityManager)
    {
        super(VolumeManagement.class, entityManager);
    }

    /**
     * Get all the volumes available to be used in persistent process and filtered by virtual
     * datacenter
     **/
    private static final String SQL_VOLUME_MANAGEMENT_GET_STATEFUL_CANDIDATES =
        "select volman.idManagement as idman, vdc.name as vdcname, rasd.limitResource as size, "
            + "rasd.reservation as available, volman.usedSize as used, "
            + "rasd.elementName as elementname, volman.state as state, tier.name as tier "
            + "from rasd, volume_management volman "
            + "left outer join rasd_management on volman.idmanagement = rasd_management.idmanagement "
            + "left outer join virtualdatacenter vdc on rasd_management.idvirtualdatacenter = vdc.idvirtualdatacenter "
            + "left outer join storage_pool pool on volman.idstorage = pool.idstorage "
            + "left outer join storage_device device on pool.idstoragedevice = device.id "
            + "left outer join tier on pool.idtier = tier.id "
            + "left outer join enterprise_limits_by_datacenter elimit on vdc.identerprise = elimit.identerprise "
            + "left outer join allowed_tier_by_enterprise alltier on elimit.iddatacenterlimit = alltier.iddatacenterlimit "
            + "where rasd_management.temporal is null "
            + "and rasd_management.idresource = rasd.instanceid "
            + "and tier.id = alltier.idtier and vdc.idvirtualdatacenter = :id_vdc "
            + "and volman.idimage is null and volman.state = :state "
            + "and device.storage_technology = :storage_tecnology";

    /** Get all volumes assigned to any VirtualDataCenter of the specified Enterprise. */
    private static final String SQL_VOLUME_MANAGEMENT_GET_VOLUMES_FROM_ENTERPRISE_AND_VDC =
        "select volman.idManagement as idman, vdc.name as vdcname, virtualapp.name as vaname, "
            + "virtualmachine.name as vmname, rasd.limitResource as size, "
            + "rasd.reservation as available, volman.usedSize as used, "
            + "rasd.elementName as elementname, volman.state as state, tier.name as tier "
            + "from (volume_management volman, virtualdatacenter vdc, rasd, rasd_management rasdm) "
            + "left join virtualmachine on rasdm.idVM = virtualmachine.idVM "
            + "left join virtualapp on rasdm.idVirtualApp = virtualapp.idVirtualApp "
            + "left join storage_pool on volman.idStorage = storage_pool.idStorage "
            + "left join tier on storage_pool.idTier = tier.id where rasdm.temporal is null "
            + "and volman.idManagement = rasdm.idManagement "
            + "and rasdm.idResource = rasd.instanceID "
            + "and rasdm.idVirtualDataCenter = vdc.idVirtualDataCenter "
            + "and vdc.idEnterprise = :idEnterprise";

    private static final String SQL_VOLUME_MANAGEMENT_GET_VOLUMES_FROM_ENTERPRISE_AND_VDC_FILTER_LIKE =
        SQL_VOLUME_MANAGEMENT_GET_VOLUMES_FROM_ENTERPRISE_AND_VDC
            + " and ( rasd.elementName like :filterLike "
            + "or virtualmachine.name like :filterLike or virtualapp.name like :filterLike "
            + "or vdc.name like :filterLike or tier.name like :filterLike )";

    private static final String SQL_VOLUME_MANAGEMENT_GET_VOLUMES_FROM_ENTERPRISE_ADD_TIER_ID_FILTER =
        SQL_VOLUME_MANAGEMENT_GET_VOLUMES_FROM_ENTERPRISE_AND_VDC + " and tier.id = :idTier";

    private static final String SQL_VOLUME_MANAGEMENT_GET_NON_PERSISTENT_VOLUMES_FROM_ENTERPRISE =
        SQL_VOLUME_MANAGEMENT_GET_VOLUMES_FROM_ENTERPRISE_AND_VDC_FILTER_LIKE
            + " AND volman.idImage IS NULL";

    private static final String SQL_RESTRICTED_VDCS_FILTER =
        " and vdc.idVirtualDataCenter IN (:vdcIds)";

    public List<VolumeManagement> getVolumesByPool(final StoragePool sp)
    {
        Criteria criteria = createCriteria(samePool(sp));
        return getResultList(criteria);
    }

    public VolumeManagement getVolumeByVirtualDatacenter(final VirtualDatacenter vdc,
        final Integer volumeId)
    {
        Criteria criteria = createCriteria(sameId(volumeId), sameVirtualDatacenter(vdc));
        return (VolumeManagement) criteria.uniqueResult();
    }

    public List<VolumeManagement> getStatefulCandidates(final VirtualDatacenter vdc)
    {

        SQLQuery query = getSession().createSQLQuery(SQL_VOLUME_MANAGEMENT_GET_STATEFUL_CANDIDATES);
        query.addScalar("idman");
        query.addScalar("vdcname");
        query.addScalar("SIZE");
        query.addScalar("available");
        query.addScalar("used");
        query.addScalar("elementname");
        query.addScalar("STATE");
        query.addScalar("tier");
        Map<String, Object> params = new HashMap<String, Object>();
        params.put("id_vdc", vdc.getId());
        params.put("state", VolumeState.DETACHED.name());
        params.put("storage_tecnology", "GENERIC_ISCSI");
        query.setProperties(params);

        return getSQLQueryResults(getSession(), query, VolumeManagement.class, 0);
    }

    @SuppressWarnings("unchecked")
    public List<VolumeManagement> getVolumesByPool(final StoragePool sp, final FilterOptions filters)
    {
        // Check if the orderBy element is actually one of the available ones
        VolumeManagement.OrderByEnum orderByEnum =
            VolumeManagement.OrderByEnum.valueOf(filters.getOrderBy().toUpperCase());

        String orderBy = defineOrderBy(orderByEnum.getColumnHQL(), filters.getAsc());

        Query query = getSession().getNamedQuery(VolumeManagement.VOLUMES_BY_POOL);

        String req = query.getQueryString() + orderBy;
        // Add order filter to the query
        Query queryWithOrder = getSession().createQuery(req);

        Map<String, String> params = new HashMap<String, String>();
        params.put("poolId", sp.getId());
        params.put("filterLike", filters.getFilter().isEmpty() ? "%" : "%" + filters.getFilter()
            + "%");
        queryWithOrder.setProperties(params);

        Integer size = count(queryWithOrder, params).intValue();

        queryWithOrder.setFirstResult(filters.getStartwith());
        queryWithOrder.setMaxResults(filters.getLimit());

        PagedList<VolumeManagement> volumesList =
            new PagedList<VolumeManagement>(queryWithOrder.list());
        volumesList.setTotalResults(size);
        volumesList.setPageSize(filters.getLimit() > size ? size : filters.getLimit());
        volumesList.setCurrentElement(filters.getStartwith());

        return volumesList;
    }

    @SuppressWarnings("unchecked")
    public List<VolumeManagement> getVolumesByVirtualDatacenter(final VirtualDatacenter vdc,
        final FilterOptions filters, final boolean includePersistent)
    {
        // Check if the orderBy element is actually one of the available ones
        VolumeManagement.OrderByEnum orderByEnum =
            VolumeManagement.OrderByEnum.valueOf(filters.getOrderBy().toUpperCase());

        String orderBy = defineOrderBy(orderByEnum.getColumnHQL(), filters.getAsc());

        Query query =
            getSession().getNamedQuery(
                includePersistent ? VolumeManagement.VOLUMES_BY_VDC
                    : VolumeManagement.VOLUMES_NON_PERSISTENT_BY_VDC);

        String req = query.getQueryString() + orderBy;
        // Add order filter to the query
        Query queryWithOrder = getSession().createQuery(req);

        Map<String, Object> params = new HashMap<String, Object>();
        params.put("vdcId", vdc.getId());
        params.put("filterLike", filters.getFilter().isEmpty() ? "%" : "%" + filters.getFilter()
            + "%");
        queryWithOrder.setProperties(params);

        Integer size = count(queryWithOrder, params).intValue();

        queryWithOrder.setFirstResult(filters.getStartwith());
        queryWithOrder.setMaxResults(filters.getLimit());

        PagedList<VolumeManagement> volumesList =
            new PagedList<VolumeManagement>(queryWithOrder.list());
        volumesList.setTotalResults(size);
        volumesList.setPageSize(filters.getLimit() > size ? size : filters.getLimit());
        volumesList.setCurrentElement(filters.getStartwith());

        return volumesList;
    }

    public VolumeManagement getVolumeByRasd(final Rasd rasd)
    {
        Criteria criteria = createCriteria(Restrictions.eq("rasd", rasd));
        return (VolumeManagement) criteria.uniqueResult();
    }

    public List<VolumeManagement> getVolumesByEnterprise(final Enterprise enterprise,
        final User currentUser, final FilterOptions filters, final boolean includePersistent)
    {
        return includePersistent ? getAllVolumesByEnterprise(enterprise, currentUser, filters)
            : getNonPersistentVolumesByEnterprise(enterprise, currentUser, filters);
    }

    public List<VolumeManagement> getAllVolumesByEnterprise(final Enterprise enterprise,
        final User currentUser, final FilterOptions filters)
    {
        // Check if the orderBy element is actually one of the available ones
        VolumeManagement.OrderByEnum orderByEnum =
            VolumeManagement.OrderByEnum.valueOf(filters.getOrderBy().toUpperCase());
        String order = defineOrderBy(orderByEnum.getColumnSQL(), filters.getAsc());
        String sql = SQL_VOLUME_MANAGEMENT_GET_VOLUMES_FROM_ENTERPRISE_AND_VDC_FILTER_LIKE;

        // Add the restricted VDC filter if the user is restricted
        Integer[] restrictedVdcs = getRestrictedVirtualDatacenters(currentUser);
        if (restrictedVdcs != null)
        {
            sql += SQL_RESTRICTED_VDCS_FILTER;
        }

        SQLQuery query = getSession().createSQLQuery(sql + order);
        addQueryAliases(query);

        Map<String, Object> params = new HashMap<String, Object>();
        params.put("idEnterprise", enterprise.getId());
        params.put("filterLike", filters.getFilter().isEmpty() ? "%" : "%" + filters.getFilter()
            + "%");
        if (restrictedVdcs != null)
        {
            params.put("vdcIds", restrictedVdcs);
        }
        query.setProperties(params);

        SQLQuery count = getSession().createSQLQuery(sql);
        count.setProperties(params);

        Integer size = countSQL(count, params).intValue();

        query.setFirstResult(filters.getStartwith());
        query.setMaxResults(filters.getLimit());
        PagedList<VolumeManagement> volumes =
            new PagedList<VolumeManagement>(getSQLQueryResults(getSession(), query,
                VolumeManagement.class, 0));
        volumes.setTotalResults(size);
        volumes.setPageSize(filters.getLimit() > size ? size : filters.getLimit());
        volumes.setCurrentElement(filters.getStartwith());

        return volumes;
    }

    public List<VolumeManagement> getNonPersistentVolumesByEnterprise(final Enterprise enterprise,
        final User currentUser, final FilterOptions filters)
    {
        // Check if the orderBy element is actually one of the available ones
        VolumeManagement.OrderByEnum orderByEnum =
            VolumeManagement.OrderByEnum.valueOf(filters.getOrderBy().toUpperCase());
        String order = defineOrderBy(orderByEnum.getColumnSQL(), filters.getAsc());
        String sql = SQL_VOLUME_MANAGEMENT_GET_NON_PERSISTENT_VOLUMES_FROM_ENTERPRISE;

        // Add the restricted VDC filter if the user is restricted
        Integer[] restrictedVdcs = getRestrictedVirtualDatacenters(currentUser);
        if (restrictedVdcs != null)
        {
            sql += SQL_RESTRICTED_VDCS_FILTER;
        }

        SQLQuery query = getSession().createSQLQuery(sql + order);
        addQueryAliases(query);

        Map<String, Object> params = new HashMap<String, Object>();
        params.put("idEnterprise", enterprise.getId());
        params.put("filterLike", filters.getFilter().isEmpty() ? "%" : "%" + filters.getFilter()
            + "%");
        if (restrictedVdcs != null)
        {
            params.put("vdcIds", restrictedVdcs);
        }
        query.setProperties(params);

        // Count without order to avoid alias issues
        SQLQuery count = getSession().createSQLQuery(sql);
        count.setProperties(params);
        Integer size = countSQL(count, params).intValue();

        query.setFirstResult(filters.getStartwith());
        query.setMaxResults(filters.getLimit());
        PagedList<VolumeManagement> volumes =
            new PagedList<VolumeManagement>(getSQLQueryResults(getSession(), query,
                VolumeManagement.class, 0));
        volumes.setTotalResults(size);
        volumes.setPageSize(filters.getLimit() > size ? size : filters.getLimit());
        volumes.setCurrentElement(filters.getStartwith());

        return volumes;
    }

    @SuppressWarnings("unchecked")
    public List<VolumeManagement> getAvailableVolumes(final VirtualDatacenter vdc,
        final FilterOptions filters)
    {
        // Check if the orderBy element is actually one of the available ones
        VolumeManagement.OrderByEnum orderByEnum =
            VolumeManagement.OrderByEnum.valueOf(filters.getOrderBy().toUpperCase());

        String orderBy = defineOrderBy(orderByEnum.getColumnHQL(), filters.getAsc());

        Query query = getSession().getNamedQuery(VolumeManagement.VOLUMES_AVAILABLES);

        String req = query.getQueryString() + orderBy;
        // Add order filter to the query
        Query queryWithOrder = getSession().createQuery(req);

        Map<String, Object> params = new HashMap<String, Object>();
        params.put("vdcId", vdc.getId());
        params.put("filterLike", filters.getFilter().isEmpty() ? "%" : "%" + filters.getFilter()
            + "%");
        queryWithOrder.setProperties(params);

        Integer size = count(queryWithOrder, params).intValue();

        // Limit 0 means no size filter
        if (filters.getLimit() == 0)
        {
            filters.setLimit(size);
            filters.setStartwith(0);
        }

        queryWithOrder.setFirstResult(filters.getStartwith());
        queryWithOrder.setMaxResults(filters.getLimit());

        PagedList<VolumeManagement> volumesList =
            new PagedList<VolumeManagement>(queryWithOrder.list());
        volumesList.setTotalResults(size);
        volumesList.setPageSize(filters.getLimit() > size ? size : filters.getLimit());
        volumesList.setCurrentElement(filters.getStartwith());

        return volumesList;
    }

    public VolumeManagement getVolumeFromImage(final Integer idImage)
    {
        Criteria criteria =
            createCriteria(
                Restrictions.eq(VolumeManagement.VIRTUAL_MACHINE_TEMPLATE_PROPERTY + "."
                    + PersistentEntity.ID_PROPERTY, idImage),
                Restrictions.isNull(RasdManagement.TEMPORAL_PROPERTY));
        Object obj = criteria.uniqueResult();
        return (VolumeManagement) obj;
    }

    @SuppressWarnings("unchecked")
    private <T> List<T> getSQLQueryResults(final Session session, final Query query,
        final Class<T> objectClass, final int idFieldPosition)
    {
        List<T> result = new ArrayList<T>();
        List<Object[]> sqlResult = query.list();

        if (sqlResult != null && !sqlResult.isEmpty())
        {
            for (Object[] res : sqlResult)
            {
                T obj = (T) session.get(objectClass, (Integer) res[idFieldPosition]);
                result.add(obj);
            }
        }

        return result;
    }

    private String defineOrderBy(final String orderBy, final Boolean asc)
    {
        StringBuilder queryString = new StringBuilder();

        queryString.append(" order by ");
        if (orderBy.equalsIgnoreCase("vol.id"))
        {
            queryString.append("vol.rasd.id");
        }
        else
        {
            queryString.append(orderBy);
        }
        queryString.append(" ");

        if (asc)
        {
            queryString.append("asc");
        }
        else
        {
            queryString.append("desc");
        }

        return queryString.toString();
    }

    private static Criterion sameVirtualDatacenter(final VirtualDatacenter vdc)
    {
        return Restrictions.eq(RasdManagement.VIRTUAL_DATACENTER_PROPERTY, vdc);
    }

    private static Criterion samePool(final StoragePool pool)
    {
        return Restrictions.eq(VolumeManagement.STORAGE_POOL_PROPERTY, pool);
    }

    private static Criterion sameId(final Integer id)
    {
        return Restrictions.eq(PersistentEntity.ID_PROPERTY, id);
    }

    @SuppressWarnings("unchecked")
    public List<VolumeManagement> getVolumesByVirtualMachine(final VirtualMachine vm,
        final FilterOptions filters)
    {
        if (filters != null)
        {
            // Check if the orderBy element is actually one of the available ones
            VolumeManagement.OrderByEnum orderByEnum =
                VolumeManagement.OrderByEnum.valueOf(filters.getOrderBy().toUpperCase());

            String orderBy = defineOrderBy(orderByEnum.getColumnHQL(), filters.getAsc());

            Query query = getSession().getNamedQuery(VolumeManagement.VOLUMES_ATTACHED_TO_VM);

            String req = query.getQueryString() + orderBy;
            // Add order filter to the query
            Query queryWithOrder = getSession().createQuery(req);

            Map<String, Object> params = new HashMap<String, Object>();
            params.put("vmId", vm.getId());
            params.put("state", VolumeState.ATTACHED);
            params.put("filterLike",
                filters.getFilter().isEmpty() ? "%" : "%" + filters.getFilter() + "%");
            queryWithOrder.setProperties(params);

            Integer size = count(queryWithOrder, params).intValue();

            // Limit 0 means no size filter
            if (filters.getLimit() == 0)
            {
                filters.setLimit(size);
                filters.setStartwith(0);
            }

            queryWithOrder.setFirstResult(filters.getStartwith());
            queryWithOrder.setMaxResults(filters.getLimit());

            PagedList<VolumeManagement> volumesList =
                new PagedList<VolumeManagement>(queryWithOrder.list());
            volumesList.setTotalResults(size);
            volumesList.setPageSize(filters.getLimit() > size ? size : filters.getLimit());
            volumesList.setCurrentElement(filters.getStartwith());

            return volumesList;
        }
        else
        {
            Criteria criteria =
                createCriteria(sameVirtualMachine(vm), sameState(VolumeState.ATTACHED));
            return getResultList(criteria);
        }
    }

    public List<VolumeManagement> getAttachedVolumes(final VirtualDatacenter vdc)
    {
        Criteria criteria =
            createCriteria(sameVirtualDatacenter(vdc), sameState(VolumeState.ATTACHED));

        return getResultList(criteria);
    }

    public List<VolumeManagement> getDetachedVolumes(final VirtualDatacenter vdc)
    {
        Criteria criteria =
            createCriteria(sameVirtualDatacenter(vdc), sameState(VolumeState.DETACHED));
        return getResultList(criteria);
    }

    private static Criterion sameVirtualMachine(final VirtualMachine vm)
    {
        return Restrictions.eq(RasdManagement.VIRTUAL_MACHINE_PROPERTY + ".id", vm.getId());
    }

    private static Criterion sameState(final VolumeState state)
    {
        return Restrictions.eq(VolumeManagement.STATE_PROPERTY, state);
    }

    private void addQueryAliases(final SQLQuery query)
    {
        query.addScalar("idman");
        query.addScalar("vdcname");
        query.addScalar("vaname");
        query.addScalar("vmname");
        query.addScalar("SIZE");
        query.addScalar("available");
        query.addScalar("used");
        query.addScalar("elementname");
        query.addScalar("STATE");
        query.addScalar("tier");
    }

    @SuppressWarnings("unchecked")
    public List<VolumeManagement> getVolumesByVirtualAppliance(final VirtualAppliance vapp)
    {
        Query query = getSession().createQuery(VolumeManagement.BY_VAPP);
        query.setInteger("vappId", vapp.getId());
        return query.list();
    }

    @Override
    public void remove(final VolumeManagement entity)
    {
        this.getEntityManager().remove(entity);
    }

    public List<VolumeManagement> getVolumesInTheTierByEnterprise(final Integer enterpriseId,
        final Integer tierId)
    {
        SQLQuery query =
            getSession().createSQLQuery(
                SQL_VOLUME_MANAGEMENT_GET_VOLUMES_FROM_ENTERPRISE_ADD_TIER_ID_FILTER);
        addQueryAliases(query);

        Map<String, Object> params = new HashMap<String, Object>();
        params.put("idEnterprise", enterpriseId);
        params.put("idTier", tierId);
        query.setProperties(params);

        return getSQLQueryResults(getSession(), query, VolumeManagement.class, 0);
    }

    private Integer[] getRestrictedVirtualDatacenters(final User user)
    {
        if (user.getAvailableVirtualDatacenters() == null)
        {
            return null;
        }
        String[] stringIds = user.getAvailableVirtualDatacenters().split(",");
        Integer[] ids = new Integer[stringIds.length];
        for (int i = 0; i < stringIds.length; i++)
        {
            ids[i] = Integer.valueOf(stringIds[i]);
        }
        return ids;
    }
}
