/**
 * Copyright (C) 2008 - Abiquo Holdings S.L. All rights reserved.
 *
 * Please see /opt/abiquo/tomcat/webapps/legal/ on Abiquo server
 * or contact contact@abiquo.com for licensing information.
 */
package com.abiquo.server.core.pricing;

import java.util.List;

import javax.persistence.EntityManager;

import org.apache.commons.lang.StringUtils;
import org.hibernate.Criteria;
import org.hibernate.criterion.Criterion;
import org.hibernate.criterion.Disjunction;
import org.hibernate.criterion.Order;
import org.hibernate.criterion.Restrictions;
import org.springframework.stereotype.Repository;

import com.abiquo.server.core.common.persistence.DefaultDAOBase;
import com.abiquo.server.core.util.PagedList;
import com.softwarementors.bzngine.entities.PersistentEntity;

@Repository("jpaPricingTemplateDAO")
public class PricingTemplateDAO extends DefaultDAOBase<Integer, PricingTemplate>
{
    public PricingTemplateDAO()
    {
        super(PricingTemplate.class);
    }

    public PricingTemplateDAO(final EntityManager entityManager)
    {
        super(PricingTemplate.class, entityManager);
    }

    public static Criterion sameName(final String name)
    {
        return Restrictions.eq(PricingTemplate.NAME_PROPERTY, name);
    }

    public boolean existAnyOtherPricingTempWithName(final PricingTemplate pt, final String name)
    {
        return existsAnyOtherByCriterions(pt, sameName(name));
    }

    private Criterion filterBy(final String filter)
    {
        Disjunction filterDisjunction = Restrictions.disjunction();

        filterDisjunction.add(Restrictions.like(PricingTemplate.NAME_PROPERTY, '%' + filter + '%'));

        return filterDisjunction;
    }

    /**
     * Look up in the DB for a pricing template with same name.
     * 
     * @param name.
     * @return boolean true if exists, false otherwise.
     */
    public boolean existAnyPricTempWithSameName(final String name)
    {
        return existsAnyByCriterions(sameName(name));
    }

    /**
     * @param startwith first element to retrieve.
     * @param numResults number of elements to retrieve.
     * @param has filter %like%
     * @param orderBy ordering filter.
     * @param asc ordering filter, ascending = true, descending = false.
     * @return List of Pricing Templates addresses that pass the filter.
     */

    public List<PricingTemplate> find(final String has, final String orderBy, final boolean desc,
        Integer numResults, Integer startwith)
    {
        Criteria criteria = createCriteria(has, orderBy, desc);

        Integer totalResults = count(criteria).intValue();

        criteria = createCriteria(has, orderBy, desc);

        // Limit 0 means no size filter
        if (numResults == 0)
        {
            numResults = totalResults;
            startwith = 0;
        }
        else if (startwith >= totalResults)
        {
            startwith = totalResults - numResults;
        }
        if (startwith != -1)
        {
            criteria.setFirstResult(startwith);
        }

        criteria.setMaxResults(numResults);

        PagedList<PricingTemplate> ptList = new PagedList<PricingTemplate>(getResultList(criteria));
        ptList.setTotalResults(totalResults);
        ptList.setPageSize(numResults);
        ptList.setCurrentElement(startwith);

        return ptList;
    }

    private Criteria createCriteria(final String filter, final String orderBy, final boolean desc)
    {
        Criteria criteria = createCriteria();

        if (!StringUtils.isEmpty(filter))
        {
            criteria.add(filterBy(filter));
        }

        if (!StringUtils.isEmpty(orderBy))
        {
            Order order = Order.asc(orderBy);
            if (desc)
            {
                order = Order.desc(orderBy);
            }
            if (!orderBy.equals(PricingTemplate.NAME_PROPERTY))
            {
                criteria.addOrder(Order.asc(PricingTemplate.CURRENCY_PROPERTY));
            }
            criteria.addOrder(order);
            criteria.addOrder(Order.asc(PricingTemplate.NAME_PROPERTY));
        }

        return criteria;
    }

    public List<PricingTemplate> findAllPricingTemplateByName(final String name)
    {
        Criteria criteria = createCriteria(sameName(name));
        criteria.addOrder(Order.asc(PricingTemplate.NAME_PROPERTY));

        return criteria.list();
    }

    public List<PricingTemplate> findPricingTemplatesByCurrency(final Integer idCurrency)
    {
        Criteria crit = createNestedCriteria(PricingTemplate.CURRENCY_PROPERTY);
        crit.add(Restrictions.eq(PersistentEntity.ID_PROPERTY, idCurrency));
        return getResultList(crit);
    }

}
