/**
 * Copyright (C) 2008 - Abiquo Holdings S.L. All rights reserved.
 *
 * Please see /opt/abiquo/tomcat/webapps/legal/ on Abiquo server
 * or contact contact@abiquo.com for licensing information.
 */
package com.abiquo.server.core.scheduler;

import java.util.Collection;
import java.util.List;

import javax.persistence.EntityManager;

import org.hibernate.Criteria;
import org.hibernate.Query;
import org.hibernate.criterion.Criterion;
import org.hibernate.criterion.Restrictions;
import org.springframework.stereotype.Repository;

import com.abiquo.server.core.common.persistence.DefaultDAOBase;
import com.abiquo.server.core.infrastructure.Machine;
import com.abiquo.server.core.infrastructure.Rack;
import com.softwarementors.bzngine.entities.PersistentEntity;

@Repository("jpaMachineLoadRuleDAO")
public class MachineLoadRuleDAO extends DefaultDAOBase<Integer, MachineLoadRule>
{
    public MachineLoadRuleDAO()
    {
        super(MachineLoadRule.class);
    }

    public MachineLoadRuleDAO(final EntityManager entityManager)
    {
        super(MachineLoadRule.class, entityManager);
    }

    private final static String CANDIDATE_MACHINE_RULES =
        "SELECT DISTINCT obj FROM com.abiquo.server.core.scheduler.MachineLoadRule obj WHERE "
            + "obj.machine IN (:machines) OR "
            + // Find rules from machines
            "obj.rack IN (SELECT obj2.rack FROM com.abiquo.server.core.infrastructure.Machine obj2 WHERE obj2 IN (:machines)) OR "
            + // Find rules from racks hosting the machines
            "obj.datacenter IN (SELECT obj3.datacenter FROM com.abiquo.server.core.infrastructure.Machine obj3 WHERE obj3 IN (:machines)) ";

    @SuppressWarnings("unchecked")
    public List<MachineLoadRule> findCandidateMachineLoadRules(
        final Collection<Machine> firstPassCandidateMachines)
    {
        Query query = getSession().createQuery(CANDIDATE_MACHINE_RULES);
        query.setParameterList("machines", firstPassCandidateMachines);

        return query.list();
    }

    public List<MachineLoadRule> getRulesForDatacenter(final Integer idDatacenter)
    {
        Criteria crit = createNestedCriteria(MachineLoadRule.DATACENTER_PROPERTY);
        crit.add(Restrictions.eq(PersistentEntity.ID_PROPERTY, idDatacenter));
        return getResultList(crit);
    }

    @SuppressWarnings("unchecked")
    public List<MachineLoadRule> findByMachine(final Integer idMachine)
    {
        Query query = getSession().createQuery(QUERY_RULES_BY_MACHINE);
        query.setInteger("idMachine", idMachine);

        List<MachineLoadRule> rules = query.list();
        return rules;
    }

    private final static String QUERY_RULES_BY_MACHINE = //
        "SELECT obj FROM com.abiquo.server.core.scheduler.MachineLoadRule obj WHERE "
            + "obj.machine.id = :idMachine";

    private final static String QUERY_RULES_BY_ALL = //
        "SELECT obj FROM com.abiquo.server.core.scheduler.MachineLoadRule obj WHERE "
            + "obj.datacenter.id = :idDatacenter and obj.rack.id = :idRack and obj.machine.id = :idMachine";

    private final static String QUERY_RULES_BY_RACK = //
        "SELECT obj FROM com.abiquo.server.core.scheduler.MachineLoadRule obj WHERE "
            + "obj.datacenter.id = :idDatacenter and obj.rack.id = :idRack and obj.machine.id is null";

    private final static String QUERY_RULES_BY_DATACENTER = //
        "SELECT obj FROM com.abiquo.server.core.scheduler.MachineLoadRule obj WHERE "
            + "obj.datacenter.id = :idDatacenter and obj.rack.id is null and obj.machine.id is null";

    public List<MachineLoadRule> findByRack(final Rack rack)
    {
        return findByRack(rack, false);
    }

    @SuppressWarnings("unchecked")
    public List<MachineLoadRule> findByRack(final Rack rack, final boolean includeMachineRules)
    {
        final Criteria criteria;
        if (includeMachineRules)
        {
            // Rules from racks and machines
            criteria = createCriteria().add(sameRack(rack));
        }
        else
        {
            // Rules from racks
            criteria = createCriteria().add(Restrictions.and(sameRack(rack), nullMachine()));
        }
        return criteria.list();
    }

    private Criterion sameRack(final Rack rack)
    {
        return Restrictions.eq(MachineLoadRule.RACK_PROPERTY, rack);
    }

    private Criterion nullMachine()
    {
        return Restrictions.isNull(MachineLoadRule.MACHINE_PROPERTY);
    }

    public List<MachineLoadRule> findMlrByMachine(final Integer datacenterId, final Integer rackId,
        final Integer machineId)
    {
        Query query = getSession().createQuery(QUERY_RULES_BY_ALL);
        query.setInteger("idDatacenter", datacenterId);
        query.setInteger("idMachine", machineId);
        query.setInteger("idRack", rackId);

        List<MachineLoadRule> rules = query.list();
        return rules;
    }

    public List<MachineLoadRule> findMlrByRack(final Integer datacenterId, final Integer rackId)
    {
        Query query = getSession().createQuery(QUERY_RULES_BY_RACK);
        query.setInteger("idDatacenter", datacenterId);
        query.setInteger("idRack", rackId);

        List<MachineLoadRule> rules = query.list();
        return rules;
    }

    public List<MachineLoadRule> findMlrByDatacenter(final Integer datacenterId)
    {
        Query query = getSession().createQuery(QUERY_RULES_BY_DATACENTER);
        query.setInteger("idDatacenter", datacenterId);

        List<MachineLoadRule> rules = query.list();
        return rules;
    }

}
