/**
 * Copyright (C) 2008 - Abiquo Holdings S.L. All rights reserved.
 *
 * Please see /opt/abiquo/tomcat/webapps/legal/ on Abiquo server
 * or contact contact@abiquo.com for licensing information.
 */
package com.abiquo.server.core.scheduler.work;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.util.UUID;

import org.hibernate.jdbc.Work;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.abiquo.server.core.cloud.Hypervisor;
import com.abiquo.server.core.cloud.VirtualMachine;
import com.abiquo.server.core.cloud.VirtualMachineState;
import com.abiquo.server.core.infrastructure.Datastore;
import com.abiquo.server.core.infrastructure.Machine;
import com.abiquo.server.core.scheduler.VirtualMachineRequirements;
import com.abiquo.server.core.task.enums.TaskType;

/**
 * <p>
 * This class executes a piece of jdbc work to write in a efficient way.
 * <p>
 * <p>
 * This class free the resources requirements of a {@link VirtualMachine} in a form of
 * {@link VirtualMachineRequirements}. This also deletes the relation between {@link Hypervisor} and
 * {@link VirtualMachine} and {@link Datastore} and {@link VirtualMachine} and
 * {@link VirtualMachine} is set to state {@link VirtualMachineState#NOT_ALLOCATED}. This work is
 * executed mainly for {@link TaskType#UNDEPLOY}.
 * </p>
 * Steps:
 * <ol>
 * <il>Delete relation {@link VirtualMachine} and {@link Datastore}</li>
 * <li>Free cpu and ram</li>
 * <li>Free hd</li>
 * <li>{@link VirtualMachine} is set to state {@link VirtualMachineState#NOT_ALLOCATED}</li>
 * </ol>
 * 
 * @author <a href="mailto:serafin.sedano@abiquo.com">Serafin Sedano</a>
 * @author <a href="mailto:daniel.estevez@abiquo.com">Daniel Estevez</a>
 */
public class FreeWork implements Work
{
    private static final Logger log = LoggerFactory.getLogger(FreeWork.class);

    private final String uuid = UUID.randomUUID().toString();

    private static final String DISSASOCIATE_VIRTUALMACHINE_SQL =
        "update virtualmachine set idHypervisor =  NULL, idDatastore = NULL, vdrpIp = NULL, vdrpPort = 0, state = 'NOT_ALLOCATED' where idVM = ?";

    private static final String FREE_RESOURCES_SQL =
        "update scheduled_resources set cpuUsed = case when cast(cpuUsed AS SIGNED) - ? > 0 then cpuUsed - ? else 0 end, ramUsed = case when cast(ramUsed AS SIGNED) - ? > 0 then ramUsed - ? else 0 end where idPhysicalMachine = ?";

    private static final String FREE_DATASTORE_HD_SQL =
        "update scheduled_resources set datastoreUsed = case when cast(datastoreUsed AS SIGNED) - ? > 0 then datastoreUsed - ? else 0 end where idDatastore = ? and idPhysicalMachine = ?";

    private VirtualMachineRequirements requirements;

    private Datastore datastore;

    private Machine machine;

    private VirtualMachine virtualMachine;

    /**
     * @param requirements The requirements to free in resources.
     * @param virtualMachine the {@link VirtualMachine} which relation must be deleted with
     *            {@link Hypervisor}.
     * @param datastore The machine (datastore) with its link to the {@link Hypervisor} from where
     *            to free resources.
     */
    public FreeWork(final VirtualMachine virtualMachine,
        final VirtualMachineRequirements requirements, final Datastore datastore)
    {
        super();
        this.requirements = requirements;
        this.virtualMachine = virtualMachine;
        this.datastore = datastore;
        this.machine = datastore.getMachine();
    }

    /**
     * Steps:
     * <ol>
     * <il>Delete relation {@link VirtualMachine}</li>
     * <li>Free cpu and ram</li>
     * <li>Free hd</li>
     * <li>{@link VirtualMachine} is set to state {@link VirtualMachineState#NOT_ALLOCATED}</li>
     * </ol>
     */
    @Override
    public void execute(final Connection connection) throws SQLException
    {
        log.trace("starting transaction work {}", uuid);
        // start the transaction by setting autocommit to false.
        // This is specially needed in mysql
        // where all statements run in their own transaction.
        connection.setAutoCommit(Boolean.FALSE);
        try
        {
            dissasociate(connection);
            freeResources(connection);
            freeDatastore(connection);
            log.trace("commit work {}...", uuid);
            connection.commit();
            log.trace("commited work {}", uuid);
        }
        catch (SQLException e)
        {
            log.error("failed to work {} due to error {}, message {}, rollback", new Object[] {
            uuid, e.getClass().getName(), e.getMessage()});
            connection.rollback();
            log.debug("failed to work {} rollback ok", uuid);
            throw e;
        }
        catch (Exception e)
        {
            log.error("failed to work {} due to error {}, message {}, rollback", new Object[] {
            uuid, e.getClass().getName(), e.getMessage()});
            connection.rollback();
            log.debug("failed to work {} rollback ok", uuid);
            throw new SQLException(e);
        }
    }

    /** Free space. */
    private void freeDatastore(final Connection connection) throws SQLException
    {

        log.trace("freeDatastore work {}...", uuid);
        Long hd = requirements.getHd();
        if (hd != null)
        {
            log.trace("work {} adding {} hd datastore        {}",
                new Object[] {uuid, hd, datastore.getName()});
            PreparedStatement statement = connection.prepareStatement(FREE_DATASTORE_HD_SQL);
            statement.setLong(1, hd);
            statement.setLong(2, hd);
            statement.setInt(3, datastore.getId());
            statement.setInt(4, machine.getId());

            statement.execute();
        }
        log.trace("freeDatastore work {} done", uuid);
    }

    /** Remove virtual machine hypervisor. */
    private void dissasociate(final Connection connection) throws SQLException
    {
        log.trace("dissasociate work {}...", uuid);
        int idVm = virtualMachine.getId();

        log.trace("work {} virtual machine {} hyperviosr {}", new Object[] {uuid, idVm});

        PreparedStatement statement = connection.prepareStatement(DISSASOCIATE_VIRTUALMACHINE_SQL);
        statement.setInt(1, idVm);

        statement.execute();
        log.trace("dissasociate work {} done", uuid);

    }

    /** Free ram and cpu. */
    private void freeResources(final Connection connection) throws SQLException
    {

        log.trace("freeResources work {}...", uuid);
        long cpu = requirements.getCpu() == null ? Long.valueOf(0) : requirements.getCpu();
        long ram = requirements.getRam() == null ? Long.valueOf(0) : requirements.getRam();
        if (cpu != 0 || ram != 0)
        {
            log.trace("work {} adding {} cpu {} ram {}", new Object[] {uuid, cpu, ram});
            PreparedStatement statement = connection.prepareStatement(FREE_RESOURCES_SQL);
            statement.setLong(1, cpu);
            statement.setLong(2, cpu);
            statement.setLong(3, ram);
            statement.setLong(4, ram);
            statement.setInt(5, machine.getId());

            statement.execute();
        }
        log.trace("freeResources work {} done", uuid);
    }

}
