/**
 * Copyright (C) 2008 - Abiquo Holdings S.L. All rights reserved.
 *
 * Please see /opt/abiquo/tomcat/webapps/legal/ on Abiquo server
 * or contact contact@abiquo.com for licensing information.
 */
package com.abiquo.server.core.scheduler.work;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.util.UUID;

import org.hibernate.jdbc.Work;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.abiquo.server.core.cloud.Hypervisor;
import com.abiquo.server.core.cloud.VirtualMachine;
import com.abiquo.server.core.infrastructure.Datastore;
import com.abiquo.server.core.infrastructure.Machine;
import com.abiquo.server.core.scheduler.VirtualMachineRequirements;
import com.abiquo.server.core.task.enums.TaskType;

/**
 * <p>
 * This class executes a piece of jdbc work to write in a efficient way.
 * <p>
 * <p>
 * This class add to the scheduled resources the requirements of a {@link VirtualMachine} in a form
 * of {@link VirtualMachineRequirements}. This work is executed mainly for {@link TaskType#DEPLOY}.
 * Also persist the relation between the {@link VirtualMachine} and the {@Hypervisor}.
 * This enables to do it in a transaction.
 * </p>
 * Steps:
 * <ol>
 * <il>Reserve cpu and ram</li>
 * <li>Reserve hd</li>
 * <li>Associate {@link Hypervisor}</li>
 * </ol>
 * 
 * @author <a href="mailto:serafin.sedano@abiquo.com">Serafin Sedano</a>
 * @author <a href="mailto:daniel.estevez@abiquo.com">Daniel Estevez</a>
 */
public class ScheduleWork implements Work
{
    private static final Logger log = LoggerFactory.getLogger(ScheduleWork.class);

    private static final String UPDATE_RESOURCES_MACHINE_SQL =
        "update scheduled_resources set cpuUsed = cpuUsed + ?, ramUsed = ramUsed + ? where idPhysicalMachine = ?";

    private static final String UPDATE_HD_DATASTORE_SQL =
        "update scheduled_resources set datastoreUsed = datastoreUsed + ? where idDatastore = ? and idPhysicalMachine = ?";

    private static final String RELATE_MACHINE_VIRTUAL_MACHINE_SQL =
        "update virtualmachine set idHypervisor = ?, idDatastore = ?, network_configuration_id = ?, vdrpIP = ?, vdrpPort = ? where idVM = ?";

    private VirtualMachineRequirements requirements;

    private Machine machine;

    private Datastore datastore;

    private VirtualMachine virtualMachine;

    private final String uuid = UUID.randomUUID().toString();

    private int vdrpPort;

    /**
     * @param virtualMachine The {@link VirtualMachine } link to the {@link Hypervisor}.
     * @param requirements The requirements to add to the scheduled resources.
     * @param datastore The machine (in datastore) with its link to the {@link Hypervisor} from
     *            where to schedule resources.
     * @param vrdpPort virtual remote desktop port.
     */
    public ScheduleWork(final VirtualMachine virtualMachine,
        final VirtualMachineRequirements requirements, final Datastore datastore, final int vrdpPort)
    {
        super();
        this.virtualMachine = virtualMachine;
        this.requirements = requirements;
        this.datastore = datastore;
        this.machine = datastore.getMachine();
        this.vdrpPort = vrdpPort;
    }

    /**
     * Steps:
     * <ol>
     * <il>Reserve cpu and ram</li>
     * <li>Reserve hd</li>
     * <li>Associate {@link Hypervisor}</li>
     * </ol>
     */
    @Override
    public void execute(final Connection connection) throws SQLException
    {
        log.trace("starting transaction work {}", uuid);
        // start the transaction by setting autocommit to false.
        // This is specially needed in mysql
        // where all statements run in their own transaction.
        connection.setAutoCommit(Boolean.FALSE);

        try
        {
            updateCpuRam(connection);
            updateHd(connection);
            associate(connection);
            log.trace("commit work {}...", uuid);
            connection.commit();
            log.trace("commited work {}", uuid);
        }
        catch (SQLException e)
        {
            log.error("failed to work {} due to error {}, message {}, rollback", new Object[] {
            uuid, e.getClass().getName(), e.getMessage()});
            connection.rollback();
            log.debug("failed to work {} rollback ok", uuid);
            throw e;
        }
        catch (Exception e)
        {
            log.error("failed to work {} due to error {}, message {}, rollback", new Object[] {
            uuid, e.getClass().getName(), e.getMessage()});
            connection.rollback();
            log.debug("failed to work {} rollback ok", uuid);
            throw new SQLException(e);
        }
    }

    private void associate(final Connection connection) throws SQLException
    {

        log.trace("associate work {}...", uuid);

        int idVm = virtualMachine.getId();
        int idHypervisor = machine.getHypervisor().getId();
        log.trace("work {} virtual machine {} hyperviosr {}", new Object[] {uuid, idVm,
        idHypervisor});

        PreparedStatement statement =
            connection.prepareStatement(RELATE_MACHINE_VIRTUAL_MACHINE_SQL);
        statement.setInt(1, idHypervisor);
        statement.setInt(2, datastore.getId());
        if (virtualMachine.getNetworkConfiguration() != null
            && virtualMachine.getNetworkConfiguration().getId() != null)
        {
            statement.setInt(3, virtualMachine.getNetworkConfiguration().getId());
        }
        else
        {
            statement.setNull(3, java.sql.Types.INTEGER);
        }
        statement.setString(4, machine.getHypervisor().getIpService());
        statement.setInt(5, vdrpPort);
        statement.setInt(6, idVm);

        statement.execute();
        log.trace("associate work {} done", uuid);
    }

    private void updateCpuRam(final Connection connection) throws SQLException
    {
        log.trace("updateCpuRam work {}...", uuid);
        long cpu = requirements.getCpu() == null ? Long.valueOf(0) : requirements.getCpu();
        long ram = requirements.getRam() == null ? Long.valueOf(0) : requirements.getRam();
        if (cpu != 0 || ram != 0)
        {
            log.trace("work {} adding {} cpu {} ram {}", new Object[] {uuid, cpu, ram});
            PreparedStatement statement = connection.prepareStatement(UPDATE_RESOURCES_MACHINE_SQL);
            statement.setLong(1, cpu);
            statement.setLong(2, ram);
            statement.setInt(3, machine.getId());

            statement.execute();
        }
        log.trace("updateCpuRam work {} done", uuid);
    }

    private void updateHd(final Connection connection) throws SQLException
    {
        log.trace("updateHd work {}...", uuid);
        Long hd = requirements.getHd();
        if (hd != null)
        {
            log.trace("work {} adding {} hd datastore {}",
                new Object[] {uuid, hd, datastore.getName()});
            PreparedStatement statement = connection.prepareStatement(UPDATE_HD_DATASTORE_SQL);
            statement.setLong(1, hd);
            statement.setInt(2, datastore.getId());
            statement.setInt(3, machine.getId());

            statement.execute();
        }
        log.trace("updateHd work {} done", uuid);
    }
}
