/**
 * Copyright (C) 2008 - Abiquo Holdings S.L. All rights reserved.
 *
 * Please see /opt/abiquo/tomcat/webapps/legal/ on Abiquo server
 * or contact contact@abiquo.com for licensing information.
 */
package com.abiquo.server.core.scheduler.work;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.util.UUID;

import org.hibernate.jdbc.Work;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.abiquo.server.core.cloud.Hypervisor;
import com.abiquo.server.core.cloud.VirtualMachine;
import com.abiquo.server.core.infrastructure.Datastore;
import com.abiquo.server.core.infrastructure.Machine;
import com.abiquo.server.core.scheduler.VirtualMachineRequirements;
import com.abiquo.server.core.task.enums.TaskType;

/**
 * <p>
 * This class executes a piece of jdbc work to write in a efficient way.
 * <p>
 * <p>
 * This class apply a patch to the resources of a {@link VirtualMachine} in a form of
 * {@link VirtualMachineRequirements}.This work is mainly executed for {@link TaskType#RECONFIGURE}.
 * </p>
 * Steps:
 * <ol>
 * <li>Update cpu and ram</li>
 * <li>Update hd</li>
 * </ol>
 * 
 * @author <a href="mailto:serafin.sedano@abiquo.com">Serafin Sedano</a>
 * @author <a href="mailto:daniel.estevez@abiquo.com">Daniel Estevez</a>
 */
public class SetResourcesWork implements Work
{
    private static final Logger log = LoggerFactory.getLogger(SetResourcesWork.class);

    private static final String UPDATE_DATASTORE_HD =
        "update scheduled_resources set datastoreUsed = ? where idDatastore = ? and idPhysicalMachine = ?";

    private final String uuid = UUID.randomUUID().toString();

    private VirtualMachineRequirements requirements;

    private Datastore datastore;


    /**
     * @param requirements The requirements to patch in resources.
     * @param datastore The machine with its link to the {@link Hypervisor} from where to patch
     *            resources.
     */
    public SetResourcesWork(final VirtualMachineRequirements requirements, final Datastore datastore)
    {
        super();
        this.requirements = requirements;
        this.datastore = datastore;
    }

    /**
     * Steps:
     * <ol>
     * <li>Update cpu and ram (not currently implemented) </li>
     * <li>Update hd</li>
     * </ol>
     */
    @Override
    public void execute(final Connection connection) throws SQLException
    {
        log.trace("starting transaction work {}", uuid);
        // start the transaction by setting autocommit to false.
        // This is specially needed in mysql
        // where all statements run in their own transaction.
        connection.setAutoCommit(Boolean.FALSE);
        try
        {
            setResources(connection);
            setDatastore(connection);

            log.trace("commit work {}...", uuid);
            connection.commit();
            log.trace("commited work {}", uuid);
        }
        catch (SQLException e)
        {
            log.error("failed to work {} due to error {}, message {}, rollback", new Object[] {
            uuid, e.getClass().getName(), e.getMessage()});
            connection.rollback();
            log.debug("failed to work {} rollback ok", uuid);
            throw e;
        }
        catch (Exception e)
        {
            log.error("failed to work {} due to error {}, message {}, rollback", new Object[] {
            uuid, e.getClass().getName(), e.getMessage()});
            connection.rollback();
            log.debug("failed to work {} rollback ok", uuid);
            throw new SQLException(e);
        }
    }

    private void setDatastore(final Connection connection) throws SQLException
    {
        log.trace("freeDatastore work {}...", uuid);
        Long hd = requirements.getHd() >= 0 ? requirements.getHd() : 0; // prevent negative usage

        if (hd != null)
        {
            log.trace("work {} setting {} hd datastore        {}", new Object[] {uuid, hd,
            datastore.getName()});
            PreparedStatement statement = connection.prepareStatement(UPDATE_DATASTORE_HD);
            statement.setLong(1, hd);
            statement.setInt(2, datastore.getId());
            statement.setInt(3, datastore.getMachine().getId());

            statement.execute();
        }
        log.trace("freeDatastore work {} done", uuid);

    }

    /**
     * Currently not implemented. Will work in the future if we can detect CPU/RAM used resources changed in machine
     * 
     * @param connection
     * @throws SQLException
     */
    private void setResources(final Connection connection) throws SQLException
    {
        // currently not implemented
    }

}
