/**
 * Copyright (C) 2008 - Abiquo Holdings S.L. All rights reserved.
 *
 * Please see /opt/abiquo/tomcat/webapps/legal/ on Abiquo server
 * or contact contact@abiquo.com for licensing information.
 */
package com.abiquo.server.core.task;

import static com.abiquo.model.redis.RedisEntityUtils.getEntityKey;
import static java.lang.Long.parseLong;
import static java.lang.String.valueOf;
import static org.apache.commons.lang.StringUtils.EMPTY;
import static org.apache.commons.lang.StringUtils.isBlank;

import java.util.HashMap;
import java.util.Map;

import org.springframework.stereotype.Component;

import redis.clients.jedis.Jedis;
import redis.clients.jedis.Transaction;

import com.abiquo.model.redis.KeyMaker;
import com.abiquo.model.redis.RedisDAOBase;
import com.abiquo.model.redis.RedisEntityUtils;
import com.abiquo.server.core.task.enums.TaskOwnerType;
import com.abiquo.server.core.task.enums.TaskState;
import com.abiquo.server.core.task.enums.TaskType;

/**
 * This base class provides Redis-persistence logic for {@link Task} entity. <h2>Instance to persist
 * </h2>
 * 
 * <pre>
 * task.ownerId = A
 * task.taskId = 0
 * task.userId = user0
 * task.type = DEPLOY
 * task.state = STARTED
 * task.timestamp = 123456789
 * task.resultType = VIRTUAL_MACHINE
 * task.resultId = A
 * task.jobs = [Job0, Job1]
 * task.creationTimestamp = 123456789
 * task.data = { "one":"somenicedata" }
 * 
 * <h2>Redis structure</h2>
 * 
 * <pre>
 * HMSET Task:0 "ownerId" "A" "taskId" "0" "userId" "user0" "type" "DEPLOY" "state" "STARTED" "timestamp" "12346789" "resultType" "VIRTUAL_MACHINE" "resultId" "A" "creationTimestamp" "12346579" "jobs" "Task:0:jobs"
 * RPUSH Task:0:jobs Jobs:0
 * RPUSH Task:0:jobs Jobs:1
 * 
 * <pre>
 * @author eruiz@abiquo.com
 */
@Component
public class TaskDAO extends RedisDAOBase<Task>
{
    private final KeyMaker keyMaker = new KeyMaker(Task.class);

    public Task findById(final String taskId, final Jedis jedis)
    {
        return find(getEntityKey(Task.class, taskId), jedis);
    }

    public Task find(final String taskKey, final Jedis jedis)
    {
        Map<String, String> hashed = jedis.hgetAll(taskKey);

        if (hashed.isEmpty())
        {
            return null;
        }

        Task task = new Task();

        task.setOwnerId(hashed.get("ownerId"));
        task.setTaskId(hashed.get("taskId"));
        task.setUserId(hashed.get("userId"));
        task.setType(TaskType.valueOf(hashed.get("type")));
        task.setState(TaskState.valueOf(hashed.get("state")));
        task.setTimestamp(parseLong(hashed.get("timestamp")));

        // For old tasks compatibility
        if (hashed.containsKey("creationTimestamp"))
        {
            task.setCreationTimestamp(parseLong(hashed.get("creationTimestamp")));
        }
        else
        {
            task.setCreationTimestamp(null);
        }

        Map<String, String> data = jedis.hgetAll(getTaskDataKey(task.getIdAsString()));

        if (data != null && !data.isEmpty())
        {
            task.getData().putAll(data);
        }

        // Set optional values
        task.setResultType(taskOwnerTypeFromString(hashed.get("resultType")));
        task.setResultId(isBlank(hashed.get("resultId")) ? null : hashed.get("resultId"));

        return task;
    }

    public TaskState findTaskState(final String taskKey, final Jedis jedis)
    {
        String state = jedis.hget(taskKey, "state");
        return state != null ? TaskState.valueOf(state) : null;
    }

    @Override
    public void delete(final Task task, final Transaction transaction)
    {
        // Build keys
        String taskJobsKey = getTaskJobsEntityKey(task.getIdAsString());
        String taskDataKey = getTaskDataKey(task.getIdAsString());

        // Delete
        transaction.del(task.getEntityKey());
        transaction.del(taskJobsKey);
        transaction.del(taskDataKey);
    }

    @Override
    public void save(final Task task, final Transaction transaction)
    {
        // Clear to persist
        delete(task, transaction);

        // Hash plain fields
        Map<String, String> hashed = new HashMap<String, String>();
        hashed.put("ownerId", task.getOwnerId());
        hashed.put("taskId", task.getTaskId());
        hashed.put("userId", task.getUserId());
        hashed.put("type", task.getType().name());
        hashed.put("state", task.getState().name());
        hashed.put("timestamp", valueOf(RedisEntityUtils.getUnixtime()));
        hashed.put("creationTimestamp", valueOf(task.getCreationTimestamp()));

        String taskJobsKey = getTaskJobsEntityKey(task.getIdAsString());
        hashed.put("jobs", taskJobsKey);

        // The following fields are optional
        String resultType = EMPTY;
        String resultId = EMPTY;

        if (task.getResultType() != null)
        {
            resultType = task.getResultType().name();
        }

        if (!isBlank(task.getResultId()))
        {
            resultId = task.getResultId();
        }

        hashed.put("resultType", resultType);
        hashed.put("resultId", resultId);

        // Hash extra data map
        String taskDataKey = getTaskDataKey(task.getIdAsString());
        hashed.put("data", taskDataKey);

        // Persist
        transaction.hmset(task.getEntityKey(), hashed);

        if (!task.getData().isEmpty())
        {
            transaction.hmset(taskDataKey, task.getData());
        }

        for (Job job : task.getJobs())
        {
            transaction.rpush(taskJobsKey, job.getEntityKey());
        }
    }

    private String getTaskDataKey(final String taskId)
    {
        return keyMaker.make(taskId, "data");
    }

    public String getTaskJobsEntityKey(final String taskId)
    {
        return keyMaker.make(taskId, "jobs");
    }

    private TaskOwnerType taskOwnerTypeFromString(final String value)
    {
        return !isBlank(value) ? TaskOwnerType.valueOf(value) : null;
    }
}
